"""Module grouping tests for the pydov.types.sondering module."""

from pydov.types.fields import ReturnFieldList
from pydov.types.sondering import Sondering
from pydov.util.dovutil import build_dov_url
from tests.abstract import AbstractTestTypes

location_wfs_getfeature = 'tests/data/types/sondering/wfsgetfeature.xml'
location_wfs_feature = 'tests/data/types/sondering/feature.xml'
location_dov_xml = 'tests/data/types/sondering/sondering.xml'


class TestSondering(AbstractTestTypes):
    """Class grouping tests for the pydov.types.sondering.Sondering class."""

    datatype_class = Sondering
    namespace = 'http://dov.vlaanderen.be/ocdov/dov-pub'
    pkey_base = build_dov_url('data/sondering/')

    field_names = [
        'pkey_sondering', 'sondeernummer', 'x', 'y', 'mv_mtaw',
        'start_sondering_mtaw', 'diepte_sondering_van',
        'diepte_sondering_tot', 'datum_aanvang', 'uitvoerder',
        'sondeermethode', 'apparaat', 'datum_gw_meting',
        'diepte_gw_m', 'lengte', 'diepte', 'qc', 'Qt', 'fs', 'u', 'i']
    field_names_subtypes = [
        'lengte', 'diepte', 'qc', 'Qt', 'fs', 'u', 'i']
    field_names_nosubtypes = [
        'pkey_sondering', 'sondeernummer', 'x', 'y', 'mv_mtaw',
        'start_sondering_mtaw', 'diepte_sondering_van',
        'diepte_sondering_tot', 'datum_aanvang', 'uitvoerder',
        'sondeermethode', 'apparaat', 'datum_gw_meting',
        'diepte_gw_m']

    valid_returnfields = ReturnFieldList.from_field_names('pkey_sondering', 'sondeernummer')
    valid_returnfields_subtype = ReturnFieldList.from_field_names('pkey_sondering', 'sondeernummer', 'z')

    inexistent_field = 'onbestaand'



"""Module grouping tests for the bodemdiepteinterval search module."""
import datetime

from owslib.fes2 import PropertyIsEqualTo

from pydov.search.bodemdiepteinterval import BodemdiepteintervalSearch
from pydov.types.bodemdiepteinterval import Bodemdiepteinterval
from pydov.types.fields import ReturnFieldList
from pydov.util.dovutil import build_dov_url
from tests.abstract import AbstractTestSearch

location_md_metadata = 'tests/data/types/bodemdiepteinterval/md_metadata.xml'
location_fc_featurecatalogue = \
    'tests/data/types/bodemdiepteinterval/fc_featurecatalogue.xml'
location_wfs_describefeaturetype = \
    'tests/data/types/bodemdiepteinterval/wfsdescribefeaturetype.xml'
location_wfs_getfeature = 'tests/data/types/bodemdiepteinterval/wfsgetfeature.xml'
location_wfs_feature = 'tests/data/types/bodemdiepteinterval/feature.xml'
location_xsd_base = 'tests/data/types/bodemdiepteinterval/xsd_*.xml'


class TestBodemdiepteintervalSearch(AbstractTestSearch):

    search_instance = BodemdiepteintervalSearch()
    datatype_class = Bodemdiepteinterval

    valid_query_single = PropertyIsEqualTo(
        propertyname='pkey_diepteinterval',
        literal=build_dov_url('data/bodemdiepteinterval/2018-000004'))

    inexistent_field = 'onbestaand'
    wfs_field = 'ondergrens_bereikt'
    xml_field = None

    valid_returnfields = ReturnFieldList.from_field_names('pkey_diepteinterval', 'naam', 'bovengrens1_cm')
    valid_returnfields_subtype = None
    valid_returnfields_extra = ReturnFieldList.from_field_names('pkey_diepteinterval', 'naam', 'Monsters')

    df_default_columns = [
        'pkey_diepteinterval', 'pkey_bodemopbouw', 'pkey_bodemlocatie',
        'nr', 'type', 'naam', 'bovengrens1_cm', 'bovengrens2_cm',
        'ondergrens1_cm', 'ondergrens2_cm', 'ondergrens_bereikt',
        'grensduidelijkheid', 'grensregelmatigheid', 'beschrijving', 'x', 'y',
        'mv_mtaw']



"""Module grouping tests for the bodemlocatie search module."""
import datetime

from owslib.fes2 import PropertyIsEqualTo

from pydov.search.bodemlocatie import BodemlocatieSearch
from pydov.types.bodemlocatie import Bodemlocatie
from pydov.types.fields import ReturnFieldList
from tests.abstract import AbstractTestSearch

location_md_metadata = 'tests/data/types/bodemlocatie/md_metadata.xml'
location_fc_featurecatalogue = \
    'tests/data/types/bodemlocatie/fc_featurecatalogue.xml'
location_wfs_describefeaturetype = \
    'tests/data/types/bodemlocatie/wfsdescribefeaturetype.xml'
location_wfs_getfeature = 'tests/data/types/bodemlocatie/wfsgetfeature.xml'
location_wfs_feature = 'tests/data/types/bodemlocatie/feature.xml'
location_dov_xml = 'tests/data/types/bodemlocatie/bodemlocatie.xml'
location_xsd_base = 'tests/data/types/bodemlocatie/xsd_*.xml'


class TestBodemlocatieSearch(AbstractTestSearch):

    search_instance = BodemlocatieSearch()
    datatype_class = Bodemlocatie

    valid_query_single = PropertyIsEqualTo(propertyname='naam',
                                           literal='STARC_4')

    inexistent_field = 'onbestaand'
    wfs_field = 'naam'
    xml_field = 'invoerdatum'

    valid_returnfields = ReturnFieldList.from_field_names('pkey_bodemlocatie', 'naam', 'waarnemingsdatum')
    valid_returnfields_subtype = ReturnFieldList.from_field_names(
        'pkey_bodemlocatie',
        'naam',
        'educatieve_waarde')
    valid_returnfields_extra = ReturnFieldList.from_field_names('pkey_bodemlocatie', 'naam', 'erfgoed')

    df_default_columns = ['pkey_bodemlocatie', 'pkey_bodemsite',
                          'naam', 'type', 'waarnemingsdatum', 'doel', 'x', 'y',
                          'mv_mtaw', 'erfgoed', 'bodemstreek',
                          'invoerdatum',
                          'educatieve_waarde']

    def test_search_date(self, mp_wfs, mp_get_schema,
                         mp_remote_describefeaturetype, mp_remote_md,
                         mp_remote_fc, mp_remote_wfs_feature, mp_dov_xml):
        """Test the search method with only the query parameter.

        Test whether the result is correct.

        Parameters
        ----------
        mp_wfs : pytest.fixture
            Monkeypatch the call to the remote GetCapabilities request.
        mp_get_schema : pytest.fixture
            Monkeypatch the call to a remote OWSLib schema.
        mp_remote_describefeaturetype : pytest.fixture
            Monkeypatch the call to a remote DescribeFeatureType.
        mp_remote_md : pytest.fixture
            Monkeypatch the call to get the remote metadata.
        mp_remote_fc : pytest.fixture
            Monkeypatch the call to get the remote feature catalogue.
        mp_remote_wfs_feature : pytest.fixture
            Monkeypatch the call to get WFS features.
        mp_dov_xml : pytest.fixture
            Monkeypatch the call to get the remote XML data.

        """
        df = self.search_instance.search(
            query=self.valid_query_single)

        # specific test for the Zulu time wfs 1.1.0 issue
        assert df.waarnemingsdatum.unique()[0] == datetime.date(2011, 9, 16)



# -*- coding: utf-8 -*-
"""Module containing the search classes to retrieve DOV bodemlocatie data."""
from ..types.bodemlocatie import Bodemlocatie
from .abstract import AbstractSearch


class BodemlocatieSearch(AbstractSearch):
    """Search class to retrieve information about bodemlocaties."""

    def __init__(self, objecttype=Bodemlocatie):
        """Initialisation.

        Parameters
        ----------
        objecttype : subclass of pydov.types.abstract.AbstractDovType
            Reference to a class representing the Bodemlocatie type.
            Optional: defaults to the Bodemlocatie type containing the fields
            described in the documentation.

        """
        super(BodemlocatieSearch, self).__init__('bodem:bodemlocaties',
                                                 objecttype)



"""Module grouping tests for returning geometry fields."""
import geopandas as gpd
from shapely.geometry import Point

from pydov.search.bodemclassificatie import BodemclassificatieSearch

location_md_metadata = 'tests/data/types/bodemclassificatie/md_metadata.xml'
location_fc_featurecatalogue = \
    'tests/data/types/bodemclassificatie/fc_featurecatalogue.xml'
location_wfs_describefeaturetype = \
    'tests/data/types/bodemclassificatie/wfsdescribefeaturetype.xml'
location_wfs_getfeature = 'tests/data/types/bodemclassificatie/wfsgetfeature.xml'
location_wfs_feature = 'tests/data/types/bodemclassificatie/feature.xml'
location_dov_xml = 'tests/data/types/bodemclassificatie/bodemclassificatie.xml'
location_xsd_base = 'tests/data/types/bodemclassificatie/xsd_*.xml'


class TestGeometryReturn(object):
    """Class grouping tests for returning geometry fields."""

    def test_return_geometry(self, mp_get_schema,
                             mp_remote_describefeaturetype,
                             mp_remote_wfs_feature):
        """Test whether the geometry field is returned when requested.

        Parameters
        ----------
        mp_get_schema : pytest.fixture
            Monkeypatch the call to a remote OWSLib schema.
        mp_remote_describefeaturetype : pytest.fixture
            Monkeypatch the call to a remote DescribeFeatureType.
        mp_remote_wfs_feature : pytest.fixture
            Monkeypatch the call to get WFS features.
        """
        return_fields = [
            'pkey_bodemclassificatie', 'pkey_bodemlocatie', 'x', 'y', 'mv_mtaw',
            'classificatietype', 'bodemtype', 'auteurs', 'geom'
        ]

        s = BodemclassificatieSearch()
        df = s.search(max_features=1, return_fields=return_fields)

        assert 'geom' in list(df)
        assert len(df.geom.notna()) == 1
        assert Point(248905.6718, 200391.287).equals_exact(df.geom[0], tolerance=0.01)

    def test_to_geopandas(self, mp_get_schema,
                          mp_remote_describefeaturetype,
                          mp_remote_wfs_feature):
        """Test whether the resulting dataframe can be turned into a GeoPandas GeoDataFrame.

        Parameters
        ----------
        mp_get_schema : pytest.fixture
            Monkeypatch the call to a remote OWSLib schema.
        mp_remote_describefeaturetype : pytest.fixture
            Monkeypatch the call to a remote DescribeFeatureType.
        mp_remote_wfs_feature : pytest.fixture
            Monkeypatch the call to get WFS features.
        """
        return_fields = [
            'pkey_bodemclassificatie', 'pkey_bodemlocatie', 'x', 'y', 'mv_mtaw',
            'classificatietype', 'bodemtype', 'auteurs', 'geom'
        ]

        s = BodemclassificatieSearch()
        df = s.search(max_features=1, return_fields=return_fields)

        geo_df = gpd.GeoDataFrame(df, geometry='geom', crs='EPSG:31370')
        assert isinstance(geo_df, gpd.GeoDataFrame)



# -*- coding: utf-8 -*-
"""Module containing the DOV data type for bodemdiepteintervallen, including
subtypes."""
from pydov.types.fields import WfsField

from .abstract import AbstractDovType


class Bodemdiepteinterval(AbstractDovType):
    """Class representing the DOV data type for bodemdiepteinterval."""

    subtypes = []

    fields = [
        WfsField(name='pkey_diepteinterval',
                 source_field='Diepteintervalfiche',
                 datatype='string'),
        WfsField(name='pkey_bodemopbouw',
                 source_field='Bodemopbouwfiche',
                 datatype='string'),
        WfsField(name='pkey_bodemlocatie',
                 source_field='Bodemlocatiefiche',
                 datatype='string'),
        WfsField(name='nr',
                 source_field='Nr',
                 datatype='integer'),
        WfsField(name='type',
                 source_field='Type',
                 datatype='string'),
        WfsField(name='naam',
                 source_field='Naam',
                 datatype='string'),
        WfsField(name='bovengrens1_cm',
                 source_field='Bovengrens1',
                 datatype='float'),
        WfsField(name='bovengrens2_cm',
                 source_field='Bovengrens2',
                 datatype='float'),
        WfsField(name='ondergrens1_cm',
                 source_field='Ondergrens1',
                 datatype='float'),
        WfsField(name='ondergrens2_cm',
                 source_field='Ondergrens2',
                 datatype='float'),
        WfsField(name='ondergrens_bereikt',
                 source_field='Ondergrens_bereikt',
                 datatype='string'),
        WfsField(name='grensduidelijkheid',
                 source_field='Grensduidelijkheid',
                 datatype='string'),
        WfsField(name='grensregelmatigheid',
                 source_field='Grensregelmatigheid',
                 datatype='string'),
        WfsField(name='beschrijving',
                 source_field='Beschrijving',
                 datatype='string'),
        WfsField(name='x',
                 source_field='X_mL72',
                 datatype='float'),
        WfsField(name='y',
                 source_field='Y_mL72',
                 datatype='float'),
        WfsField(name='mv_mtaw',
                 source_field='mv_mTAW',
                 datatype='float'),
    ]

    pkey_fieldname = 'Diepteintervalfiche'

    def __init__(self, pkey):
        """Initialisation.

        Parameters
        ----------
        pkey : str
            Permanent key of the Bodemdiepteinterval, being a URI of the form
            `https://www.dov.vlaanderen.be/data/bodemdiepteinterval/<id>`.

        """
        super().__init__('bodemdiepteinterval', pkey)



# -*- coding: utf-8 -*-
"""Module containing the search classes to retrieve DOV bodemclassificatie
data."""
from ..types.bodemclassificatie import Bodemclassificatie
from .abstract import AbstractSearch


class BodemclassificatieSearch(AbstractSearch):
    """Search class to retrieve information about bodemclassificaties."""

    def __init__(self, objecttype=Bodemclassificatie):
        """Initialisation.

        Parameters
        ----------
        objecttype : subclass of pydov.types.abstract.AbstractDovType
            Reference to a class representing the Bodemclassificatie type.
            Optional: defaults to the Bodemclassificatie type containing the
            fields described in the documentation.

        """
        super(BodemclassificatieSearch, self).__init__(
            'bodem:bodemclassificaties', objecttype)



"""Module grouping tests for the pydov.types.boring module."""

from pydov.types.fields import ReturnFieldList
from pydov.types.grondwatervergunning import GrondwaterVergunning
from tests.abstract import AbstractTestTypes

location_wfs_getfeature = \
    'tests/data/types/grondwatervergunning/wfsgetfeature.xml'
location_wfs_feature = 'tests/data/types/grondwatervergunning/feature.xml'
location_dov_xml = None


class TestGrondwaterVergunning(AbstractTestTypes):
    """Class grouping tests for the
    pydov.types.grondwatervergunning.GrondwaterVergunning class."""

    datatype_class = GrondwaterVergunning
    namespace = 'http://dov.vlaanderen.be/grondwater/gw_vergunningen'
    pkey_base = None

    field_names = [
        'id_vergunning', 'pkey_installatie', 'x', 'y',
        'diepte', 'exploitant_naam', 'watnr', 'vlaremrubriek',
        'vergund_jaardebiet', 'vergund_dagdebiet',
        'van_datum_termijn', 'tot_datum_termijn',
        'aquifer_vergunning', 'inrichtingsklasse', 'nacebelcode',
        'actie_waakgebied', 'cbbnr', 'kbonr']
    field_names_subtypes = None
    field_names_nosubtypes = [
        'id_vergunning', 'pkey_installatie', 'x', 'y',
        'diepte', 'exploitant_naam', 'watnr', 'vlaremrubriek',
        'vergund_jaardebiet', 'vergund_dagdebiet',
        'van_datum_termijn', 'tot_datum_termijn',
        'aquifer_vergunning', 'inrichtingsklasse', 'nacebelcode',
        'actie_waakgebied', 'cbbnr', 'kbonr']

    valid_returnfields = ReturnFieldList.from_field_names('id_vergunning', 'diepte')
    valid_returnfields_subtype = None

    inexistent_field = 'onbestaand'



"""Tests regarding dates and timestamps."""

import pytest
import numpy as np

from tests.test_search_grondwatervergunning import (
    location_md_metadata, location_fc_featurecatalogue,
    location_wfs_describefeaturetype, location_xsd_base
)

from pydov.search.grondwatervergunning import GrondwaterVergunningSearch
from pydov.util.errors import DataParseWarning

location_wfs_getfeature = 'tests/data/date/wfsgetfeature.xml'
location_wfs_feature = 'tests/data/date/feature.xml'


class TestSearchInvalidDate(object):
    """Test the behaviour when encoutering invalid dates in the data."""

    def test_year_gt_9999(self, mp_wfs, mp_get_schema,
                          mp_remote_describefeaturetype, mp_remote_md,
                          mp_remote_fc, mp_remote_wfs_feature):
        """Test with a year greater than 9999. This is not supported in Python.

        Test whether no exception is raised, but instead the date is replaced with NaN and a warning is raised.

        Parameters
        ----------
        mp_wfs : pytest.fixture
            Monkeypatch the call to the remote GetCapabilities request.
        mp_get_schema : pytest.fixture
            Monkeypatch the call to a remote OWSLib schema.
        mp_remote_describefeaturetype : pytest.fixture
            Monkeypatch the call to a remote DescribeFeatureType.
        mp_remote_md : pytest.fixture
            Monkeypatch the call to get the remote metadata.
        mp_remote_fc : pytest.fixture
            Monkeypatch the call to get the remote feature catalogue.
        mp_remote_wfs_feature : pytest.fixture
            Monkeypatch the call to get WFS features.
        """
        s = GrondwaterVergunningSearch()

        with pytest.warns(DataParseWarning):
            df = s.search(max_features=1)
            assert np.isnan(df.tot_datum_termijn.iloc[0])



# -*- coding: utf-8 -*-
"""Module containing the DOV data type for screens (Filter), including
subtypes."""
from pydov.types.fields import WfsField, XmlField, XsdType
from pydov.types.ligging import MvMtawField
from pydov.util.dovutil import build_dov_url

from .abstract import AbstractDovSubType, AbstractDovType

_filterDataCodes_xsd = build_dov_url(
    'xdov/schema/latest/xsd/kern/gwmeetnet/FilterDataCodes.xsd')


class Peilmeting(AbstractDovSubType):

    rootpath = './/filtermeting/peilmeting'

    fields = [
        XmlField(name='datum',
                 source_xpath='/datum',
                 definition='Datum waarop de peilmeting uitgevoerd werd.',
                 datatype='date'),
        XmlField(name='tijdstip',
                 source_xpath='/tijdstip',
                 definition='Tijdstip waarop de peilmeting uitgevoerd werd ('
                            'optioneel).',
                 datatype='string'),
        XmlField(name='peil_mtaw',
                 source_xpath='/peil_mtaw',
                 definition='Diepte van de peilmeting, uitgedrukt in mTAW.',
                 datatype='float'),
        XmlField(name='betrouwbaarheid',
                 source_xpath='/betrouwbaarheid',
                 definition='Betrouwbaarheid van de peilmeting (goed, '
                            'onbekend of twijfelachtig).',
                 datatype='string'),
        XmlField(name='methode',
                 source_xpath='/methode',
                 definition='Methode waarop de peilmeting uitgevoerd werd.',
                 datatype='string',
                 xsd_type=XsdType(
                     xsd_schema=_filterDataCodes_xsd,
                     typename='PeilmetingMethodeEnumType')),
        XmlField(name='filterstatus',
                 source_xpath='/filterstatus',
                 definition='Status van de filter tijdens de peilmeting (in '
                            'rust - werking).',
                 datatype='string',
                 xsd_type=XsdType(
                     xsd_schema=_filterDataCodes_xsd,
                     typename='FilterstatusEnumType')),
        XmlField(name='filtertoestand',
                 source_xpath='/filtertoestand',
                 definition="Filtertoestand bij de peilmeting. "
                            "Standaardwaarde is '1' = Normaal.",
                 datatype='integer',
                 xsd_type=XsdType(
                     xsd_schema=_filterDataCodes_xsd,
                     typename='FiltertoestandEnumType'))
    ]


class GrondwaterFilter(AbstractDovType):
    """Class representing the DOV data type for Groundwater screens."""

    subtypes = [Peilmeting]

    fields = [
        WfsField(name='pkey_filter', source_field='filterfiche',
                 datatype='string'),
        WfsField(name='pkey_grondwaterlocatie', source_field='putfiche',
                 datatype='string'),
        WfsField(name='gw_id', source_field='GW_ID', datatype='string'),
        WfsField(name='filternummer', source_field='filternummer',
                 datatype='string'),
        WfsField(name='filtertype', source_field='filtertype',
                 datatype='string'),
        WfsField(name='x', source_field='X_mL72', datatype='float'),
        WfsField(name='y', source_field='Y_mL72', datatype='float'),
        WfsField(name='start_grondwaterlocatie_mtaw', source_field='Z_mTAW',
                 datatype='float'),
        MvMtawField('Maaiveldhoogte in mTAW op dag '
                    'dat de put/boring uitgevoerd werd'),
        WfsField(name='gemeente', source_field='gemeente', datatype='string'),
        XmlField(name='meetnet_code',
                 source_xpath='/filter/meetnet',
                 definition='Tot welk meetnet behoort deze filter.',
                 datatype='string',
                 xsd_type=XsdType(
                     xsd_schema=_filterDataCodes_xsd,
                     typename='MeetnetEnumType')),
        XmlField(name='aquifer_code',
                 source_xpath='/filter/ligging/aquifer',
                 definition='De aquifer (watervoerende laag) waarin de filter '
                            'hangt (code) (HCOVv1)',
                 datatype='string',
                 xsd_type=XsdType(
                     xsd_schema=build_dov_url(
                         'xdov/schema/latest/xsd/kern/interpretatie/'
                         'HydrogeologischeStratigrafieDataCodes.xsd'),
                     typename='AquiferHCOVv1EnumType')),
        XmlField(name='grondwaterlichaam_code',
                 source_xpath='/filter/ligging/grondwaterlichaam',
                 definition='',
                 datatype='string',
                 xsd_type=XsdType(
                     xsd_schema=_filterDataCodes_xsd,
                     typename='GrondwaterlichaamEnumType')),
        XmlField(name='regime',
                 source_xpath='/filter/ligging/regime',
                 definition='',
                 datatype='string'),
        WfsField(name='diepte_onderkant_filter',
                 source_field='onderkant_filter_m', datatype='float'),
        WfsField(name='lengte_filter', source_field='lengte_filter_m',
                 datatype='float')
    ]

    pkey_fieldname = 'filterfiche'

    def __init__(self, pkey):
        """Initialisation.

        Parameters
        ----------
        pkey : str
            Permanent key of the Filter (screen), being a URI of the form
            `https://www.dov.vlaanderen.be/data/filter/<id>`.

        """
        super().__init__('filter', pkey)



# -*- coding: utf-8 -*-
import pydov.util.caching
from pydov.util.hooks import Hooks, SimpleStatusHook
from pydov.util.net import SessionFactory, proxy_autoconfiguration

__author__ = """DOV-Vlaanderen"""
__version__ = '3.2.0'

cache = pydov.util.caching.GzipTextFileCache()

hooks = Hooks(
    (SimpleStatusHook(),)
)

# try proxy autoconfiguration via PAC
proxy_autoconfiguration()

# Package wide requests session object. This increases performance as using a
# session object allows connection pooling and TCP connection reuse.
session = SessionFactory.get_session()



# -*- coding: utf-8 -*-
"""Module containing the search classes to retrieve DOV interpretations."""

from pydov.search.abstract import AbstractSearch
from pydov.types.interpretaties import (FormeleStratigrafie,
                                        GecodeerdeLithologie,
                                        GeotechnischeCodering,
                                        HydrogeologischeStratigrafie,
                                        InformeleHydrogeologischeStratigrafie,
                                        InformeleStratigrafie,
                                        LithologischeBeschrijvingen,
                                        QuartairStratigrafie)


class InformeleStratigrafieSearch(AbstractSearch):
    """Search class to retrieve information about 'informele stratigrafie'."""

    def __init__(self, objecttype=InformeleStratigrafie):
        """Initialisation.

        Parameters
        ----------
        objecttype : subclass of pydov.types.abstract.AbstractDovType
            Reference to a class representing the InformeleStratigrafie type.
            Optional: defaults to the InformeleStratigrafie type
            containing the fields described in the documentation.

        """
        super(InformeleStratigrafieSearch, self).__init__(
            'interpretaties:informele_stratigrafie', objecttype)


class FormeleStratigrafieSearch(AbstractSearch):
    """Search class to retrieve the interpretation for Formele
    stratigrafie"""

    def __init__(self, objecttype=FormeleStratigrafie):
        """Initialisation.

        Parameters
        ----------
        objecttype : subclass of pydov.types.abstract.AbstractDovType
            Reference to a class representing the FormeleStratigrafie type.
            Optional: defaults to the FormeleStratigrafie type containing the
            fields described in the documentation.

        """
        super(FormeleStratigrafieSearch, self).__init__(
            'interpretaties:formele_stratigrafie', objecttype)


class HydrogeologischeStratigrafieSearch(AbstractSearch):
    """Search class to retrieve hydrogeological interpretations """

    def __init__(self, objecttype=HydrogeologischeStratigrafie):
        """Initialisation.

        Parameters
        ----------
        objecttype : subclass of pydov.types.abstract.AbstractDovType
            Reference to a class representing the HydrogeologischeStratigrafie
            type. Optional: defaults to the HydrogeologischeStratigrafie type
            containing the fields described in the documentation.

        """
        super(HydrogeologischeStratigrafieSearch, self).__init__(
            'interpretaties:hydrogeologische_stratigrafie',
            objecttype)


class LithologischeBeschrijvingenSearch(AbstractSearch):
    """Search class to retrieve lithologische beschrijvingen """

    def __init__(self, objecttype=LithologischeBeschrijvingen):
        """Initialisation.

        Parameters
        ----------
        objecttype : subclass of pydov.types.abstract.AbstractDovType
            Reference to a class representing the LithologischeBeschrijvingen
            type. Optional: defaults to the LithologischeBeschrijvingen type
            containing the fields described in the documentation.

        """
        super(LithologischeBeschrijvingenSearch, self).__init__(
            'interpretaties:lithologische_beschrijvingen',
            objecttype)


class GecodeerdeLithologieSearch(AbstractSearch):
    """Search class to retrieve gecodeerde lithologie """

    def __init__(self, objecttype=GecodeerdeLithologie):
        """Initialisation.

        Parameters
        ----------
        objecttype : subclass of pydov.types.abstract.AbstractDovType
            Reference to a class representing the GecodeerdeLithologie type.
            Optional: defaults to the GecodeerdeLithologie type containing
            the fields described in the documentation.

        """
        super(GecodeerdeLithologieSearch, self).__init__(
            'interpretaties:gecodeerde_lithologie',
            objecttype)


class GeotechnischeCoderingSearch(AbstractSearch):
    """Search class to retrieve geotechnische codering """

    def __init__(self, objecttype=GeotechnischeCodering):
        """Initialisation.

        Parameters
        ----------
        objecttype : subclass of pydov.types.abstract.AbstractDovType
            Reference to a class representing the GeotechnischeCodering type.
            Optional: defaults to the GeotechnischeCodering type containing
            the fields described in the documentation.

        """
        super(GeotechnischeCoderingSearch, self).__init__(
            'interpretaties:geotechnische_coderingen',
            objecttype)


class QuartairStratigrafieSearch(AbstractSearch):
    """Search class to retrieve the interpretation for Quartair
    stratigrafie"""

    def __init__(self, objecttype=QuartairStratigrafie):
        """Initialisation.

        Parameters
        ----------
        objecttype : subclass of pydov.types.abstract.AbstractDovType
            Reference to a class representing the QuartairStratigrafie type.
            Optional: defaults to the QuartairStratigrafie type containing
            the fields described in the documentation.

        """
        super(QuartairStratigrafieSearch, self).__init__(
            'interpretaties:quartaire_stratigrafie', objecttype)


class InformeleHydrogeologischeStratigrafieSearch(AbstractSearch):
    """Search class to retrieve information about informele
    hydrogeologische stratigrafie.

    Parameters
    ----------
    objecttype : subclass of pydov.types.abstract.AbstractDovType
        Reference to a class representing the
        InformeleHydrogeologischeStratigrafie type.
        Optional: defaults to the InformeleHydrogeologischeStratigrafie type
        containing the fields described in the documentation.

    """

    def __init__(self, objecttype=InformeleHydrogeologischeStratigrafie):
        """Initialisation."""
        super(InformeleHydrogeologischeStratigrafieSearch, self).__init__(
            'interpretaties:informele_hydrogeologische_stratigrafie',
            objecttype)



# -*- coding: utf-8 -*-
"""Module grouping Exception classes."""


class DOVError(Exception):
    """General error within PyDOV."""
    pass


class RemoteFetchError(DOVError):
    """General error while fetching remote data from DOV webservices."""
    pass


class OWSError(DOVError):
    """Error regarding the OGC web services."""
    pass


class LayerNotFoundError(OWSError):
    """Error that occurs when a specific layer could not be found."""
    pass


class WfsGetFeatureError(OWSError):
    """Error that occurs when requesting features using WFS GetFeature."""
    pass


class InvalidSearchParameterError(DOVError):
    """Error that occurs when given invalid parameters to the DOV search."""
    pass


class InvalidFieldError(DOVError):
    """Error that occurs when using a field outside its scope."""
    pass


class XmlParseError(DOVError):
    """Error that occurs when the parsing of a DOV XML document failed."""
    pass


class LogReplayError(Exception):
    pass


class DOVWarning(Warning):
    """General warning in PyDOV."""
    pass


class XsdFetchWarning(DOVWarning):
    """Emitted when an XSD document fails to be fetched from the DOV
    webservice, resulting in the fields metadata to be incomplete."""


class XmlFetchWarning(DOVWarning):
    """Emitted when an XML document fails to be fetched from the DOV
    webservice, resulting in an incomplete dataframe."""


class XmlStaleWarning(DOVWarning):
    """Emitted when an XML document fails to be fetched from the DOV
    webservice and an older stale version is used from the cache, resulting
    in an out-of-date dataframe."""


class XmlParseWarning(DOVWarning):
    """Emitted when the failure to parse an XML document results in
    an incomplete dataframe."""


class DataParseWarning(DOVWarning):
    """Emitted when some data retrieved from DOV failed to be parsed according
    to the datatypes in pydov. The data that failed to be parsed is removed,
    resulting in an incomplete dataframe."""



import datetime
import random
import re
from collections import OrderedDict

import numpy as np
import pandas as pd
import pytest
import requests
from owslib.etree import etree
from owslib.fes2 import PropertyIsEqualTo, SortBy, SortProperty
from pandas import DataFrame

import pydov
from pydov.types.abstract import AbstractField
from pydov.types.fields import ReturnField, ReturnFieldList
from pydov.util.dovutil import build_dov_url
from pydov.util.errors import InvalidFieldError
from pydov.util.location import Box, Within
from pydov.util.query import Join, PropertyInList


class ServiceCheck:

    is_service_ok = None

    @staticmethod
    def service_ok(timeout=5):
        """Check whether DOV services are accessible.

        Used to skip online tests when the service is unavailable or
        unreachable.

        Parameters
        ----------
        timeout : int, optional
            Timeout in seconds. Defaults to 5.

        Returns
        -------
        bool
            True if the DOV services are reachable, False otherwise.

        """
        def check_url(url, timeout):
            try:
                ok = pydov.session.head(
                    url, allow_redirects=True, timeout=timeout).ok
            except requests.exceptions.ReadTimeout:
                ok = False
            except requests.exceptions.ConnectTimeout:
                ok = False
            except Exception:
                ok = False
            return ok

        if ServiceCheck.is_service_ok is None:
            ServiceCheck.is_service_ok = (
                check_url(build_dov_url('geonetwork'), timeout) and
                check_url(build_dov_url('geoserver'), timeout) and
                check_url(
                    build_dov_url('xdov/schema/latest/xsd/kern/dov.xsd'),
                    timeout))

        return ServiceCheck.is_service_ok


def clean_xml(xml):
    """Clean the given XML string of namespace definition, namespace
    prefixes and syntactical but otherwise meaningless differences.

    Parameters
    ----------
    xml : str
        String representation of XML document.

    Returns
    -------
    str
        String representation of cleaned XML document.

    """
    # remove xmlns namespace definitions
    r = re.sub(r'[ ]+xmlns:[^=]+="[^"]+"', '', xml)

    # remove namespace prefixes in tags
    r = re.sub(r'<(/?)[^:]+:([^ >]+)([ >])', r'<\1\2\3', r)

    # remove namespace prefixes in attributes
    while re.match(r'<([^ >]+)( [^:]+ )*[^:]+:([^ >]+)([ >])', r):
        r = re.sub(r'<([^ >]+)( [^:]+ )*[^:]+:([^ >]+)([ >])', r'<\1\2\3\4', r)

    # remove extra spaces in tags
    r = re.sub(r'[ ]+/>', '/>', r)

    # remove extra spaces between tags
    r = re.sub(r'>[ ]+<', '><', r)

    return r


class AbstractTestSearch(object):
    """Class grouping common test code for search classes.

    Subclasses should implement at least the following public attributes
    in order for the tests defined here to be executed.

    Attributes
    ----------
    search_instance : pydov.search.abstract.AbstractSearch
        Instance of subclass of this type used for searching.
    datatype_class : pydov.types.abstract.AbstractDovType
            Class reference for the corresponding datatype.
    valid_query_single : owslib.fes2.OgcExpression
        OGC expression of a valid query returning a single result.
    inexistent_field : str
            The name of an inexistent field.
    wfs_field : str
        The name of a WFS field.
    xml_field : str
        The name of an XML field.
    valid_returnfields : tuple of str
        A tuple of valid return fields from the main type.
    valid_returnfields_subtype : typle of str
            A tuple containing valid return fields, including fields from a
            subtype.
    valid_returnfields_extra : tuple of str
            A tuple containing valid return fields, including extra fields
            from WFS, not present in the default dataframe.
    df_default_columns : list of str
            A list of the column names of the default dataframe.

    """

    def test_pluggable_type(self):
        """Test whether the search object can be initialised by explicitly
        giving the objecttype.
        """
        datatype = self.datatype_class
        self.search_instance.__class__(objecttype=datatype)

    def test_get_fields(self, mp_wfs, mp_get_schema,
                        mp_remote_describefeaturetype, mp_remote_md,
                        mp_remote_fc, mp_remote_xsd):
        """Test the get_fields method.

        Test whether the returned fields match the format specified
        in the documentation.

        Parameters
        ----------
        mp_wfs : pytest.fixture
            Monkeypatch the call to the remote GetCapabilities request.
        mp_get_schema : pytest.fixture
            Monkeypatch the call to a remote OWSLib schema.
        mp_remote_describefeaturetype : pytest.fixture
            Monkeypatch the call to a remote DescribeFeatureType.
        mp_remote_md : pytest.fixture
            Monkeypatch the call to get the remote metadata.
        mp_remote_fc : pytest.fixture
            Monkeypatch the call to get the remote feature catalogue.

        """
        fields = self.search_instance.get_fields()

        assert isinstance(fields, dict)

        for field in fields:
            assert isinstance(field, str)

            f = fields[field]
            assert isinstance(f, dict)

            assert 'name' in f
            assert isinstance(f['name'], str)
            assert f['name'] == field

            assert 'definition' in f
            assert isinstance(f['name'], str)

            assert 'type' in f
            assert isinstance(f['type'], str)
            assert f['type'] in ['string', 'float', 'integer', 'date',
                                 'datetime', 'boolean', 'geometry']

            assert 'notnull' in f
            assert isinstance(f['notnull'], bool)

            assert 'query' in f
            assert isinstance(f['query'], bool)

            assert 'cost' in f
            assert isinstance(f['cost'], int)
            assert f['cost'] > 0

            if 'values' in f:
                assert sorted(f.keys()) == [
                    'cost', 'definition', 'name', 'notnull', 'query', 'type',
                    'values']

                assert isinstance(f['values'], dict)

                for v in f['values'].keys():
                    assert isinstance(f['values'][v], str) or f[
                        'values'][v] is None

                    if f['type'] == 'string':
                        assert isinstance(v, str)
                    elif f['type'] == 'float':
                        assert isinstance(v, float)
                    elif f['type'] == 'integer':
                        assert isinstance(v, int)
                    elif f['type'] == 'date':
                        assert isinstance(v, datetime.date)
                    elif f['type'] == 'datetime':
                        assert isinstance(v, datetime.datetime)
                    elif f['type'] == 'boolean':
                        assert isinstance(v, bool)
            else:
                assert sorted(f.keys()) == ['cost', 'definition', 'name',
                                            'notnull', 'query', 'type']

    def test_search_both_location_query(self, mp_get_schema,
                                        mp_remote_describefeaturetype,
                                        mp_remote_wfs_feature):
        """Test the search method providing both a location and a query.

        Test whether a dataframe is returned.

        Parameters
        ----------
        mp_get_schema : pytest.fixture
            Monkeypatch the call to a remote OWSLib schema.
        mp_remote_describefeaturetype : pytest.fixture
            Monkeypatch the call to a remote DescribeFeatureType.
        mp_remote_wfs_feature : pytest.fixture
            Monkeypatch the call to get WFS features.

        """
        df = self.search_instance.search(
            location=Within(Box(1, 2, 3, 4)),
            query=self.valid_query_single,
            return_fields=self.valid_returnfields)

        assert isinstance(df, DataFrame)

    def test_search(self, mp_wfs, mp_get_schema, mp_remote_describefeaturetype,
                    mp_remote_md, mp_remote_fc, mp_remote_wfs_feature,
                    mp_dov_xml):
        """Test the search method with only the query parameter.

        Test whether the result is correct.

        Parameters
        ----------
        mp_wfs : pytest.fixture
            Monkeypatch the call to the remote GetCapabilities request.
        mp_get_schema : pytest.fixture
            Monkeypatch the call to a remote OWSLib schema.
        mp_remote_describefeaturetype : pytest.fixture
            Monkeypatch the call to a remote DescribeFeatureType.
        mp_remote_md : pytest.fixture
            Monkeypatch the call to get the remote metadata.
        mp_remote_fc : pytest.fixture
            Monkeypatch the call to get the remote feature catalogue.
        mp_remote_wfs_feature : pytest.fixture
            Monkeypatch the call to get WFS features.
        mp_dov_xml : pytest.fixture
            Monkeypatch the call to get the remote XML data.

        """
        df = self.search_instance.search(
            query=self.valid_query_single)

        assert isinstance(df, DataFrame)

        assert list(df) == self.df_default_columns

        datatype = self.datatype_class
        allfields = datatype.get_field_names()
        ownfields = datatype.get_field_names(include_subtypes=False)
        subfields = [f for f in allfields if f not in ownfields]

        assert len(df) >= 1

        for field in list(df):
            if field in ownfields:
                assert len(df[field].unique()) == 1
            elif field in subfields:
                assert len(df[field].unique()) >= 1

        # dtype checks of the resulting df columns
        fields = self.search_instance.get_fields()

        for field in list(df):
            mandatory = fields[field]['notnull']
            if mandatory:
                assert len(df[field]) == len(df[field].dropna())

        for field in list(df):
            field_datatype = fields[field]['type']
            datatypes = set((type(i) for i in df[field].dropna()))

            assert len(datatypes) <= 1

            if len(datatypes) > 0:
                if field_datatype == 'string':
                    assert (str in datatypes)
                elif field_datatype == 'float':
                    assert float in datatypes
                elif field_datatype == 'integer':
                    assert (int in datatypes)
                elif field_datatype == 'date':
                    assert datetime.date in datatypes
                elif field_datatype == 'boolean':
                    assert bool in datatypes

    def test_search_returnfields(self, mp_remote_wfs_feature):
        """Test the search method with the query parameter and a selection of
        return fields.

        Test whether the output dataframe contains only the selected return
        fields.

        Parameters
        ----------
        mp_remote_wfs_feature : pytest.fixture
            Monkeypatch the call to get WFS features.

        """
        df = self.search_instance.search(
            query=self.valid_query_single,
            return_fields=self.valid_returnfields)

        assert isinstance(df, DataFrame)

        assert list(df) == list(self.valid_returnfields.get_names())

    def test_search_returnfields_subtype(self, mp_remote_wfs_feature):
        """Test the search method with the query parameter and a selection of
        return fields, including fields from a subtype.

        Test whether the output dataframe contains only the selected return
        fields.

        Parameters
        ----------
        mp_remote_wfs_feature : pytest.fixture
            Monkeypatch the call to get WFS features.

        """
        if self.valid_returnfields_subtype is None:
            return

        df = self.search_instance.search(
            query=self.valid_query_single,
            return_fields=self.valid_returnfields_subtype)

        assert isinstance(df, DataFrame)

        assert list(df) == list(self.valid_returnfields_subtype.get_names())

    def test_search_returnfields_order(self, mp_remote_wfs_feature):
        """Test the search method with the query parameter and a selection of
        return fields in another ordering.

        Test whether the output dataframe contains only the selected return
        fields, in the order that is given in the return_fields parameter.

        Parameters
        ----------
        mp_remote_wfs_feature : pytest.fixture
            Monkeypatch the call to get WFS features.

        """
        rf = list(self.valid_returnfields.get_names())

        while rf == list(self.valid_returnfields):
            random.shuffle(rf)

        df = self.search_instance.search(
            query=self.valid_query_single,
            return_fields=rf)

        assert isinstance(df, DataFrame)
        assert list(df) == rf

    def test_search_wrongreturnfields(self):
        """Test the search method with the query parameter and an inexistent
        return field.

        Test whether an InvalidFieldError is raised.

        """
        return_fields = list(self.valid_returnfields)
        return_fields.append(self.inexistent_field)

        with pytest.raises(InvalidFieldError):
            self.search_instance.search(
                query=self.valid_query_single,
                return_fields=return_fields)

    def test_search_wrongreturnfieldstype(self):
        """Test the search method with the query parameter and a single
        return field as string.

        Test whether an AttributeError is raised.

        """
        with pytest.raises(AttributeError):
            self.search_instance.search(
                query=self.valid_query_single,
                return_fields=self.valid_returnfields[0])

    def test_search_query_wrongfield(self):
        """Test the search method with the query parameter using an
        inexistent query field.

        Test whether an InvalidFieldError is raised.

        """
        query = PropertyIsEqualTo(propertyname=self.inexistent_field,
                                  literal='The cat is out of the bag.')

        with pytest.raises(InvalidFieldError):
            self.search_instance.search(
                query=query)

    def test_search_query_wrongfield_returnfield(self):
        """Test the search method with the query parameter using an
        return-only field as query field.

        Test whether an InvalidFieldError is raised.

        """
        if self.xml_field is None:
            return

        query = PropertyIsEqualTo(propertyname=self.xml_field,
                                  literal='Geotechnisch onderzoek')

        with pytest.raises(InvalidFieldError):
            self.search_instance.search(query=query)

    def test_search_extrareturnfields(self, mp_get_schema,
                                      mp_remote_describefeaturetype,
                                      mp_remote_wfs_feature, mp_dov_xml):
        """Test the search method with the query parameter and an extra WFS
        field as return field.

        Parameters
        ----------
        mp_get_schema : pytest.fixture
            Monkeypatch the call to a remote OWSLib schema.
        mp_remote_describefeaturetype : pytest.fixture
            Monkeypatch the call to a remote DescribeFeatureType.
        mp_remote_wfs_feature : pytest.fixture
            Monkeypatch the call to get WFS features.
        mp_dov_xml : pytest.fixture
            Monkeypatch the call to get the remote XML data.

        """
        if self.valid_returnfields_extra is None:
            return

        df = self.search_instance.search(
            query=self.valid_query_single,
            return_fields=self.valid_returnfields_extra)

        assert isinstance(df, DataFrame)

        assert list(df) == list(self.valid_returnfields_extra.get_names())

    @pytest.mark.online
    @pytest.mark.skipif(not ServiceCheck.service_ok(),
                        reason="DOV service is unreachable")
    def test_search_sortby_valid(self):
        """Test the search method with the query parameter and the sort_by
        parameter with a valid sort field.

        Test whether a dataframe is returned.
        """
        df = self.search_instance.search(
            query=self.valid_query_single,
            sort_by=SortBy([SortProperty(
                self.valid_returnfields.get_names()[-1])]))

        assert isinstance(df, DataFrame)

    def test_search_sortby_invalid(self, mp_get_schema,
                                   mp_remote_describefeaturetype,
                                   mp_remote_wfs_feature, mp_dov_xml):
        """Test the search method with the query parameter and the sort_by
        parameter with an invalid sort field.

        Test whether an InvalidFieldError is raised.

        Parameters
        ----------
        mp_get_schema : pytest.fixture
            Monkeypatch the call to a remote OWSLib schema.
        mp_remote_describefeaturetype : pytest.fixture
            Monkeypatch the call to a remote DescribeFeatureType.
        mp_remote_wfs_feature : pytest.fixture
            Monkeypatch the call to get WFS features.
        mp_dov_xml : pytest.fixture
            Monkeypatch the call to get the remote XML data.

        """
        if self.xml_field is None:
            return

        with pytest.raises(InvalidFieldError):
            self.search_instance.search(
                query=self.valid_query_single,
                sort_by=SortBy([SortProperty(
                    self.xml_field)]))

    def test_search_xml_noresolve(self, mp_get_schema,
                                  mp_remote_describefeaturetype,
                                  mp_remote_wfs_feature, mp_dov_xml_broken):
        """Test the search method with return fields from WFS only.

        Test whether no XML is resolved.

        Parameters
        ----------
        mp_get_schema : pytest.fixture
            Monkeypatch the call to a remote OWSLib schema.
        mp_remote_describefeaturetype : pytest.fixture
            Monkeypatch the call to a remote DescribeFeatureType.
        mp_remote_wfs_feature : pytest.fixture
            Monkeypatch the call to get WFS features.
        mp_dov_xml_broken : pytest.fixture
            Monkeypatch the call to break fetching of remote XML data.

        """
        self.search_instance.search(
            query=self.valid_query_single,
            return_fields=self.valid_returnfields_extra)

    def test_search_propertyinlist(self, mp_get_schema,
                                   mp_remote_describefeaturetype,
                                   mp_remote_wfs_feature, mp_dov_xml):
        """Test the search method with a PropertyInList query.

        Parameters
        ----------
        mp_get_schema : pytest.fixture
            Monkeypatch the call to a remote OWSLib schema.
        mp_remote_describefeaturetype : pytest.fixture
            Monkeypatch the call to a remote DescribeFeatureType.
        mp_remote_wfs_feature : pytest.fixture
            Monkeypatch the call to get WFS features.
        mp_dov_xml : pytest.fixture
            Monkeypatch the call to get the remote XML data.

        """
        self.search_instance.search(
            query=PropertyInList(self.wfs_field, ['a', 'b']))

    def test_search_join(self, mp_get_schema, mp_remote_describefeaturetype,
                         mp_remote_wfs_feature, mp_dov_xml):
        """Test the search method with a Join query.

        Parameters
        ----------
        mp_get_schema : pytest.fixture
            Monkeypatch the call to a remote OWSLib schema.
        mp_remote_describefeaturetype : pytest.fixture
            Monkeypatch the call to a remote DescribeFeatureType.
        mp_remote_wfs_feature : pytest.fixture
            Monkeypatch the call to get WFS features.
        mp_dov_xml : pytest.fixture
            Monkeypatch the call to get the remote XML data.

        """
        df1 = self.search_instance.search(
            query=self.valid_query_single)

        self.search_instance.search(
            query=Join(df1, self.df_default_columns[0]))

    def test_get_fields_xsd_values(self, mp_remote_xsd):
        """Test the result of get_fields when the XML field has an XSD type.

        Test whether the output from get_fields() returns the values from
        the XSD.

        Parameters
        ----------
        mp_remote_xsd : pytest.fixture
            Monkeypatch the call to get XSD schemas.

        """
        xsd_schemas = self.datatype_class.get_xsd_schemas()

        if len(xsd_schemas) > 0:
            xml_fields = self.datatype_class.get_fields(source='xml')
            fields = self.search_instance.get_fields()
            for f in xml_fields.values():
                if 'xsd_type' in f:
                    assert 'values' in fields[f['name']]
                    assert isinstance(fields[f['name']]['values'], dict)

    def test_get_fields_no_xsd(self):
        """Test whether no XML fields have an XSD type when no XSD schemas
        are available."""
        xsd_schemas = self.datatype_class.get_xsd_schemas()

        if len(xsd_schemas) == 0:
            xml_fields = self.datatype_class.get_fields(source='xml')
            for f in xml_fields.values():
                assert 'xsd_type' not in f

    def test_get_fields_xsd_enums(self):
        """Test whether at least one XML field has an XSD type when there
        are XSD schemas available.

        Make sure XSD schemas are only listed (and downloaded) when they are
        needed.

        """
        xsd_schemas = self.datatype_class.get_xsd_schemas()

        xsd_type_count = 0

        if len(xsd_schemas) > 0:
            xml_fields = self.datatype_class.get_fields(source='xml')
            for f in xml_fields.values():
                if 'xsd_type' in f:
                    xsd_type_count += 1
            assert xsd_type_count > 0


class AbstractTestTypes(object):
    """Class grouping common test code for datatype classes.

    Subclasses should implement at least the following public attributes
    in order for the tests defined here to be executed.

    Attributes
    ----------
    datatype_class : pydov.types.abstract.AbstractDovType
        Class reference for the corresponding datatype.
    namespace : str
        WFS namespace for this type.
    pkey_base : str
        Base URL for the permanent keys of this datatype.
    field_names : list of str
        The field names for this type as listed in the documentation in
        docs/description_output_dataframes.rst
    field_names_subtypes : list of str
        The field names of this type that originate from subtypes only.
    field_names_nosubtypes : list of str
        The field names for this type, without including fields from
        subtypes.
    valid_returnfields : tuple of str
        A tuple of valid return fields from the main type.
    valid_returnfields_subtype : typle of str
            A tuple containing valid return fields, including fields from a
            subtype.
    inexistent_field : str
            The name of an inexistent field.

    """

    def test_get_field_names(self):
        """Test the get_field_names method.

        Tests whether the available fields for the type match the
        ones we list in docs/description_output_dataframes.rst.

        """
        fields = self.datatype_class.get_field_names()
        assert fields == self.field_names

    def test_get_field_names_nosubtypes(self):
        """Test the get_field_names method without including subtypes.

        Tests whether the fields provided in a subtype are not listed when
        disabling subtypes.

        """
        fields = self.datatype_class.get_field_names(
            return_fields=None, include_subtypes=False)

        assert fields == self.field_names_nosubtypes

    def test_get_field_names_returnfields_nosubtypes(self):
        """Test the get_field_names method when specifying return
        fields.

        Tests whether the returned fields match the ones provided as return
        fields.

        """
        fields = self.datatype_class.get_field_names(
            return_fields=self.valid_returnfields,
            include_subtypes=False)

        assert fields == list(self.valid_returnfields.get_names())

    def test_get_field_names_returnfields_order(self):
        """Test the get_field_names method when specifying return
        fields in a different order.

        Tests whether the returned fields match the ones provided as return
        fields and that the order is the one given in the return_fields
        parameter.

        """
        rf = ReturnFieldList(self.valid_returnfields)

        while rf == self.valid_returnfields:
            random.shuffle(rf)

        fields = self.datatype_class.get_field_names(
            return_fields=rf,
            include_subtypes=False)

        assert fields == rf.get_names()

    def test_get_field_names_wrongreturnfields(self):
        """Test the get_field_names method when specifying an
        inexistent return field.

        Test whether an InvalidFieldError is raised.

        """
        return_fields = self.valid_returnfields
        return_fields.append(ReturnField(self.inexistent_field))

        with pytest.raises(InvalidFieldError):
            self.datatype_class.get_field_names(
                return_fields=return_fields,
                include_subtypes=False)

    def test_get_field_names_wrongreturnfieldstype(self):
        """Test the get_field_names method when listing a single
        return field instead of a list.

        Test whether an AttributeError is raised.

        """
        with pytest.raises(AttributeError):
            self.datatype_class.get_field_names(
                return_fields=self.valid_returnfields.get_names()[0],
                include_subtypes=False)

    def test_get_field_names_wrongreturnfields_nosubtypes(self):
        """Test the get_field_names method when disabling subtypes
        and including an otherwise valid return field.

        Test whether an InvalidFieldError is raised.

        """
        if self.valid_returnfields_subtype is None:
            return

        with pytest.raises(InvalidFieldError):
            self.datatype_class.get_field_names(
                return_fields=self.valid_returnfields_subtype,
                include_subtypes=False)

    def test_get_fields(self):
        """Test the get_fields method.

        Test whether the format returned by the method meets the
        requirements and the format listed in the docs.

        """
        fields = self.datatype_class.get_fields()

        assert isinstance(fields, OrderedDict)

        for f in fields.keys():
            assert isinstance(f, str)

            field = fields[f]
            assert isinstance(field, AbstractField)

            assert 'name' in field
            assert isinstance(field['name'], str)
            assert field['name'] == f

            assert 'source' in field
            assert isinstance(field['source'], str)
            assert field['source'] in ('wfs', 'xml')

            assert 'sourcefield' in field
            assert isinstance(field['sourcefield'], str)

            assert 'type' in field
            assert isinstance(field['type'], str)
            assert field['type'] in ['string', 'float', 'integer', 'date',
                                     'datetime', 'boolean', 'geometry']

            if field['source'] == 'wfs':
                if 'wfs_injected' in field.keys():
                    assert sorted(field.keys()) == [
                        'name', 'source', 'sourcefield', 'type',
                        'wfs_injected']
                else:
                    assert sorted(field.keys()) == [
                        'name', 'source', 'sourcefield', 'type']
            elif field['source'] == 'xml':
                assert 'definition' in field
                assert isinstance(field['definition'], str)

                assert 'notnull' in field
                assert isinstance(field['notnull'], bool)

                if 'xsd_type' in field:
                    assert sorted(field.keys()) == [
                        'definition', 'name', 'notnull', 'source',
                        'sourcefield', 'type', 'xsd_schema', 'xsd_type']
                else:
                    assert sorted(field.keys()) == [
                        'definition', 'name', 'notnull', 'source',
                        'sourcefield', 'type']

    def test_get_fields_nosubtypes(self):
        """Test the get_fields method not including subtypes.

        Test whether fields provides by subtypes are not listed in the output.

        """
        if self.field_names_subtypes is None:
            return

        fields = self.datatype_class.get_fields(include_subtypes=False)
        for field in fields:
            assert field not in self.field_names_subtypes

    def test_from_wfs_element(self, wfs_feature):
        """Test the from_wfs_element method.

        Test whether we can construct an instance from a WFS response element.

        Parameters
        ----------
        wfs_feature : pytest.fixture returning etree.Element
            Fixture providing an XML element representing a single record of
            the WFS layer.

        """
        feature = self.datatype_class.from_wfs_element(
            wfs_feature, self.namespace)

        assert isinstance(feature, self.datatype_class)

        if self.pkey_base is not None:
            assert feature.pkey.startswith(self.pkey_base)

            assert feature.pkey.startswith(
                build_dov_url('data/{}/'.format(feature.typename)))

        assert isinstance(feature.data, dict)
        assert isinstance(feature.subdata, dict)

    def test_get_df_array(self, wfs_feature, mp_dov_xml):
        """Test the get_df_array method.

        Test whether the output of the dataframe array is correct.

        Parameters
        ----------
        wfs_feature : pytest.fixture returning etree.Element
            Fixture providing an XML element representing a single record of
            the WFS layer.
        mp_dov_xml : pytest.fixture
            Monkeypatch the call to get the remote XML data.

        """
        feature = self.datatype_class.from_wfs_element(
            wfs_feature, self.namespace)

        fields = [f for f in self.datatype_class.get_fields(
            source=('wfs', 'xml', 'custom_wfs', 'custom_xml')).values() if not
            f.get('wfs_injected', False)]

        df_array = feature.get_df_array()

        assert isinstance(df_array, list)

        for record in df_array:
            assert len(record) == len(fields)

            for value, field in zip(record, fields):
                if field['type'] == 'string':
                    assert isinstance(value, str) or np.isnan(value)
                elif field['type'] == 'float':
                    assert isinstance(value, float) or np.isnan(value)
                elif field['type'] == 'integer':
                    assert isinstance(value, int) or np.isnan(value)
                elif field['type'] == 'date':
                    assert isinstance(value, datetime.date) or np.isnan(value)
                elif field['type'] == 'boolean':
                    assert isinstance(value, bool) or np.isnan(value)

                if field['name'].startswith('pkey') and not pd.isnull(value):
                    assert value.startswith(build_dov_url('data/'))
                    assert not value.endswith('.xml')

    def test_get_df_array_wrongreturnfields(self, wfs_feature):
        """Test the get_df_array specifying a nonexistent return field.

        Test whether an InvalidFieldError is raised.

        Parameters
        ----------
        wfs_feature : pytest.fixture returning etree.Element
            Fixture providing an XML element representing a single record of
            the WFS layer.

        """
        feature = self.datatype_class.from_wfs_element(
            wfs_feature, self.namespace)

        with pytest.raises(InvalidFieldError):
            feature.get_df_array(
                return_fields=ReturnFieldList.from_field_names(self.inexistent_field))

    def test_from_wfs_str(self, wfs_getfeature):
        """Test the from_wfs method to construct objects from a WFS response,
        as str.

        Parameters
        ----------
        wfs_getfeature : pytest.fixture returning str
            Fixture providing a WFS GetFeature response.

        """
        features = self.datatype_class.from_wfs(wfs_getfeature,
                                                self.namespace)

        for feature in features:
            assert isinstance(feature, self.datatype_class)

    def test_from_wfs_bytes(self, wfs_getfeature):
        """Test the from_wfs method to construct objects from a WFS response,
        as bytes.

        Parameters
        ----------
        wfs_getfeature : pytest.fixture returning str
            Fixture providing a WFS GetFeature response.

        """
        features = self.datatype_class.from_wfs(wfs_getfeature.encode('utf-8'),
                                                self.namespace)

        for feature in features:
            assert isinstance(feature, self.datatype_class)

    def test_from_wfs_tree(self, wfs_getfeature):
        """Test the from_wfs method to construct objects from a WFS response,
        as elementtree.

        Parameters
        ----------
        wfs_getfeature : pytest.fixture returning str
            Fixture providing a WFS GetFeature response.

        """
        tree = etree.fromstring(wfs_getfeature.encode('utf8'))
        features = self.datatype_class.from_wfs(tree, self.namespace)

        for feature in features:
            assert isinstance(feature, self.datatype_class)

    def test_from_wfs_list(self, wfs_getfeature):
        """Test the from_wfs method to construct objects from a WFS response,
        as list of elements.

        Parameters
        ----------
        wfs_getfeature : pytest.fixture returning str
            Fixture providing a WFS GetFeature response.

        """
        tree = etree.fromstring(wfs_getfeature.encode('utf8'))
        feature_members = tree.find(
            './/{http://www.opengis.net/wfs/2.0}member')

        if feature_members is not None:
            fts = [ft for ft in feature_members]

            features = self.datatype_class.from_wfs(fts, self.namespace)

            for feature in features:
                assert isinstance(feature, self.datatype_class)

    def test_missing_pkey(self):
        """Test initialising an object type with a pkey of 'None'.

        Test whether a ValueError is raised.

        """
        with pytest.raises(ValueError):
            self.datatype_class(None)

    def test_nested_subtype_from_xml_element(self, dov_xml):
        """Test initialising the subtype(s) from the XML document.

        Parameters
        ----------
        dov_xml : pytest.fixture returning bytes
            Fixture providing DOV XML data.
        """
        def instance_from_xml(clz, xml):
            if len(clz.subtypes) == 0:
                return

            if xml is None:
                return

            st_instance = next(
                clz.subtypes[0].from_xml(dov_xml))
            assert isinstance(st_instance, clz.subtypes[0])

            instance_from_xml(clz.subtypes[0], dov_xml)

        instance_from_xml(self.datatype_class, dov_xml)



# -*- coding: utf-8 -*-
"""Module containing the DOV data type for CPT measurements (Sonderingen),
including subtypes."""
from pydov.types.abstract import AbstractDovSubType, AbstractDovType
from pydov.types.fields import WfsField, XmlField
from pydov.types.ligging import MvMtawField


class Meetdata(AbstractDovSubType):

    rootpath = './/sondering/sondeonderzoek/penetratietest/meetdata'

    fields = [
        XmlField(name='lengte',
                 source_xpath='/lengte',
                 definition='Geregistreerde sondeerlengte, '
                            'uitgedrukt in meter.',
                 datatype='float'),
        XmlField(name='diepte',
                 source_xpath='/diepte',
                 definition='Diepte waarop sondeerparameters geregistreerd '
                            'werden, berekend uit de sondeerlengte en de '
                            'geregistreerde hellingsmeting, '
                            'uitgedrukt in meter.',
                 datatype='float'),
        XmlField(name='qc',
                 source_xpath='/qc',
                 definition='Opgemeten waarde van de conusweerstand, '
                            'uitgedrukt in MPa.',
                 datatype='float'),
        XmlField(name='Qt',
                 source_xpath='/Qt',
                 definition='Opgemeten waarde van de totale weerstand, '
                            'uitgedrukt in kN.',
                 datatype='float'),
        XmlField(name='fs',
                 source_xpath='/fs',
                 definition='Opgemeten waarde van de plaatelijke '
                            'kleefweerstand, uitgedrukt in kPa.',
                 datatype='float'),
        XmlField(name='u',
                 source_xpath='/u',
                 definition='Opgemeten waarde van de porienwaterspanning, '
                            'uitgedrukt in kPa.',
                 datatype='float'),
        XmlField(name='i',
                 source_xpath='/i',
                 definition='Opgemeten waarde van de inclinatie, uitgedrukt '
                            'in graden.',
                 datatype='float')
    ]


class Sondering(AbstractDovType):
    """Class representing the DOV data type for CPT measurements."""

    subtypes = [Meetdata]

    fields = [
        WfsField(name='pkey_sondering', source_field='fiche',
                 datatype='string'),
        WfsField(name='sondeernummer', source_field='sondeernummer',
                 datatype='string'),
        WfsField(name='x', source_field='X_mL72', datatype='float'),
        WfsField(name='y', source_field='Y_mL72', datatype='float'),
        MvMtawField('Maaiveldhoogte in mTAW op dag dat de sondering '
                    'uitgevoerd werd.'),
        WfsField(name='start_sondering_mtaw', source_field='Z_mTAW',
                 datatype='float'),
        WfsField(name='diepte_sondering_van', source_field='diepte_van_m',
                 datatype='float'),
        WfsField(name='diepte_sondering_tot', source_field='diepte_tot_m',
                 datatype='float'),
        WfsField(name='datum_aanvang', source_field='datum_aanvang',
                 datatype='date'),
        WfsField(name='uitvoerder', source_field='uitvoerder',
                 datatype='string'),
        WfsField(name='sondeermethode', source_field='sondeermethode',
                 datatype='string'),
        WfsField(name='apparaat', source_field='apparaat_type',
                 datatype='string'),
        XmlField(name='datum_gw_meting',
                 source_xpath='/sondering/visueelonderzoek/'
                              'datumtijd_waarneming_grondwaterstand',
                 definition='Datum en tijdstip van waarneming van de '
                            'grondwaterstand.',
                 datatype='datetime'),
        XmlField(name='diepte_gw_m',
                 source_xpath='/sondering/visueelonderzoek/grondwaterstand',
                 definition='Diepte water in meter ten opzicht van het '
                            'aanvangspeil.',
                 datatype='float')
    ]

    pkey_fieldname = 'fiche'

    def __init__(self, pkey):
        """Initialisation.

        Parameters
        ----------
        pkey : str
            Permanent key of the Sondering (CPT measurement), being a URI of
            the form `https://www.dov.vlaanderen.be/data/sondering/<id>`.

        """
        super().__init__('sondering', pkey)



# -*- coding: utf-8 -*-
"""Module implementing a simple hooks system to allow late-binding actions to
PyDOV events."""

from hashlib import md5
import importlib
from multiprocessing import Lock
from owslib.etree import etree
from pathlib import Path
import atexit
import json
import math
import os
import sys
import time
import uuid
import zipfile

import pydov
from pydov.util.errors import LogReplayError


class Hooks(list):
    """Runtime representation of registered pydov hooks, i.e. a list of
    instances of AbstractReadHook and/or AbstractInjectHook."""

    def get_read_hooks(self):
        """Get the registered read hooks (i.e. hooks that are subclasses of
        AbstractReadHook), in the order they are defined in the list.

        Returns
        -------
        tuple of AbstractReadHook
            A tuple with the registered read hooks.

        """
        return (h for h in self if isinstance(h, AbstractReadHook))

    def get_inject_hooks(self):
        """Get the registered inject hooks (i.e. hooks that are subclasses of
        AbstractInjectHook), in the order they are defined in the list.

        Returns
        -------
        tuple of AbstractInjectHook
            A tuple with the registered inject hooks.

        """
        return (h for h in self if isinstance(h, AbstractInjectHook))


class HookRunner(object):
    """Class for executing registered hooks."""
    @staticmethod
    def __execute_read(hook_name, args):
        """Execute the read hook with the given name for all registered hooks.

        Parameters
        ----------
        hook_name : str
            Name of the hook function to execute.
        args : list
            List of arguments to pass to the hook function call.

        """
        for h in pydov.hooks.get_read_hooks():
            getattr(h, hook_name)(*args)

    @staticmethod
    def __execute_inject(hook_name, args):
        """Execute the inject hook with given name for all registered hooks
        and return the last non-null result.

        Parameters
        ----------
        hook_name : str
            Name of the hook function to execute.
        args : list
            List of arguments to pass to the hook function call.

        Returns
        -------
        object or bytes or None
            Returns the last non-null result of the execution of the inject
            hook. If all inject hooks return None, return None as well.

        """
        result = None
        for h in pydov.hooks.get_inject_hooks():
            r = getattr(h, hook_name)(*args)
            if r is not None:
                result = r
        return result

    @staticmethod
    def execute_meta_received(url, response):
        """Execute the meta_received method for all registered hooks.

        Parameters
        ----------
        url : str
            URL of the metadata request.
        response : bytes
            The raw response as received from resolving the URL.

        """
        HookRunner.__execute_read('meta_received', [url, response])

    @staticmethod
    def execute_wfs_search_init(params):
        """Execute the wfs_search_init method for all registered hooks.

        Parameters
        ----------
        params : dict
            Parameters used to initiate WFS search. These include:
                typename : str
                    Typename in the WFS service to query.
                location : pydov.util.location.AbstractLocationFilter or None
                    Location filter limiting the features to retrieve.
                filter : str (xml) or None
                    Attribute filter limiting the features to retrieve.
                sort_by : str (xml) or None
                    SortBy clause listing fields to sort by.
                max_features : int
                    Limit the maximum number of features to request.
                propertynames : list of str
                    List of WFS propertynames (attributes) to retrieve.
                geometry_column : str
                    Name of the column/attribute containing the geometry.

        """
        HookRunner.__execute_read('wfs_search_init', [params])

    @staticmethod
    def execute_wfs_search_result(number_matched, number_returned):
        """Execute the wfs_search_result method for all registered hooks.

        Parameters
        ----------
        number_matched : int
            The number of features matched by the WFS search query.
        number_returned : int
            The number of features returned by the WFS search query. Due to
            server limitations this can be less than number_matched.

        """
        HookRunner.__execute_read(
            'wfs_search_result', [number_matched, number_returned])

    @staticmethod
    def execute_wfs_search_result_received(query, features):
        """Execute the wfs_search_result_received method for all registered
        hooks.

        Parameters
        ----------
        query : etree.ElementTree
            The WFS GetFeature request sent to the WFS server.
        features : etree.ElementTree
            The WFS GetFeature response containing the features.

        """
        HookRunner.__execute_read('wfs_search_result_received', [
            query, features])

    @staticmethod
    def execute_xml_received(pkey_object, xml):
        """Execute the xml_received method for all registered hooks.

        Parameters
        ----------
        pkey_object : str
            Permanent key of the retrieved object.
        xml : bytes
            The raw XML data of this DOV object as bytes.

        """
        HookRunner.__execute_read('xml_received', [pkey_object, xml])

    @staticmethod
    def execute_xml_cache_hit(pkey_object):
        """Execute the xml_cache_hit method for all registered hooks.

        Parameters
        ----------
        pkey_object : str
            Permanent key of the requested object.
        """
        HookRunner.__execute_read('xml_cache_hit', [pkey_object])

    @staticmethod
    def execute_xml_stale_hit(pkey_object):
        """Execute the xml_stale_hit method for all registered hooks.

        Parameters
        ----------
        pkey_object : str
            Permanent key of the requested object.

        """
        HookRunner.__execute_read('xml_stale_hit', [pkey_object])

    @staticmethod
    def execute_xml_fetch_error(pkey_object):
        """Execute the xml_fetch_error method for all registered hooks.

        Parameters
        ----------
        pkey_object : str
            Permanent key of the requested object.

        """
        HookRunner.__execute_read('xml_fetch_error', [pkey_object])

    @staticmethod
    def execute_xml_downloaded(pkey_object):
        """Execute the xml_downloaded method for all registered hooks.

        Parameters
        ----------
        pkey_object : str
            Permanent key of the requested object.

        """
        HookRunner.__execute_read('xml_downloaded', [pkey_object])

    @staticmethod
    def execute_inject_meta_response(url):
        """Execute the inject_meta_response method for all registered hooks.

        Parameters
        ----------
        url : str
            URL of the metadata request.

        Returns
        -------
        bytes, optional
            The response to use in favor of resolving the URL. Returns None if
            this inject hook is unused.

        """
        return HookRunner.__execute_inject(
            'inject_meta_response', [url])

    @staticmethod
    def execute_inject_wfs_getfeature_response(query):
        """Execute the inject_wfs_getfeature_response method for all
        registered hooks.

        Parameters
        ----------
        query : etree.ElementTree
            The WFS GetFeature request sent to the WFS server.

        Returns
        -------
        xml: bytes, optional
            The GetFeature response to use in favor of resolving the URL.
            Returns None if this inject hook is unused.

        """
        return HookRunner.__execute_inject(
            'inject_wfs_getfeature_response', [query])

    @staticmethod
    def execute_inject_xml_response(pkey_object):
        """Execute the inject_xml_response method for all registered hooks.

        Parameters
        ----------
        query : etree.ElementTree
            The WFS GetFeature request sent to the WFS server.

        Returns
        -------
        xml : bytes, optional
            The XML response to use in favor of resolving the URL. Returns
            None if this inject hook is unused.

        """
        return HookRunner.__execute_inject(
            'inject_xml_response', [pkey_object])


class AbstractReadHook(object):
    """Abstract base class for custom hook implementations.

    This class contains all read-only hooks: i.e. hooks receiving events but
    otherwise not interfering with pydov's execution stack.

    Provides all available methods with a default implementation to do
    nothing. This allows for hook subclasses to only implement the events
    they need.

    """

    def meta_received(self, url, response):
        """Called when a response for a metadata requests is received.

        Metadata calls include amongst others: WFS GetCapabilities, requests
        for MD_Metadata, FC_FeatureCatalogue and XSD schemas.

        These are all calls except for WFS GetFeature requests and XML
        downloads of DOV data - these are other hooks.

        Parameters
        ----------
        url : str
            URL of the metadata request.
        response : bytes
            The raw response as received from resolving the URL.

        """
        pass

    def wfs_search_init(self, params):
        """Called upon starting a WFS search.

        Parameters
        ----------
        params : dict
            Parameters used to initiate WFS search. These include:
                typename : str
                    Typename in the WFS service to query.
                location : pydov.util.location.AbstractLocationFilter or None
                    Location filter limiting the features to retrieve.
                filter : str (xml) or None
                    Attribute filter limiting the features to retrieve.
                sort_by : str (xml) or None
                    SortBy clause listing fields to sort by.
                max_features : int
                    Limit the maximum number of features to request.
                propertynames : list of str
                    List of WFS propertynames (attributes) to retrieve.
                geometry_column : str
                    Name of the column/attribute containing the geometry.

        """
        pass

    def wfs_search_result(self, number_matched, number_returned):
        """Called after a WFS search query finished.

        Because of parallel processing, this method will be called
        simultaneously from multiple threads. Make sure your implementation is
        threadsafe or uses locking.

        Parameters
        ----------
        number_matched : int
            The number of features matched by the WFS search query.
        number_returned : int
            The number of features returned by the WFS search query. Due to
            server limitations this can be less than number_matched.

        """
        pass

    def wfs_search_result_received(self, query, features):
        """Called after a WFS search finished.

        Includes both the GetFeature query as well as the response from the
        WFS server.

        Because of parallel processing, this method will be called
        simultaneously from multiple threads. Make sure your implementation is
        threadsafe or uses locking.

        Parameters
        ----------
        query : etree.ElementTree
            The WFS GetFeature request sent to the WFS server.
        features : etree.ElementTree
            The WFS GetFeature response containings the features.

        """
        pass

    def xml_received(self, pkey_object, xml):
        """Called when the XML of a given object is received, either from
        the cache or from the remote DOV service.

        Includes the permanent key of the DOV object as well as the full XML
        representation.

        Because of parallel processing, this method will be called
        simultaneously from multiple threads. Make sure your implementation is
        threadsafe or uses locking.

        Parameters
        ----------
        pkey_object : str
            Permanent key of the retrieved object.
        xml : bytes
            The raw XML data of this DOV object as bytes.

        """
        pass

    def xml_cache_hit(self, pkey_object):
        """Called when the XML document of an object is retrieved from the
        cache.

        Because of parallel processing, this method will be called
        simultaneously from multiple threads. Make sure your implementation is
        threadsafe or uses locking.

        Parameters
        ----------
        pkey_object : str
            Permanent key of the requested object.

        """
        pass

    def xml_stale_hit(self, pkey_object):
        """Called when the XML document of an object failed to be retrieved
        from the DOV service and a stale version has been returned from the
        cache.

        Because of parallel processing, this method will be called
        simultaneously from multiple threads. Make sure your implementation is
        threadsafe or uses locking.

        Parameters
        ----------
        pkey_object : str
            Permanent key of the requested object.

        """
        pass

    def xml_fetch_error(self, pkey_object):
        """Called when the XML document of an object failed to be retrieved
        from the DOV service and no stale version could be returned from the
        cache.

        Because of parallel processing, this method will be called
        simultaneously from multiple threads. Make sure your implementation is
        threadsafe or uses locking.

        Parameters
        ----------
        pkey_object : str
            Permanent key of the requested object.

        """
        pass

    def xml_downloaded(self, pkey_object):
        """Called when the XML document of an object is downloaded from the
        DOV services.

        Because of parallel processing, this method will be called
        simultaneously from multiple threads. Make sure your implementation is
        threadsafe or uses locking.

        Parameters
        ----------
        pkey_object : str
            Permanent key of the requested object.

        """
        pass


class AbstractInjectHook(object):
    """Abstract base class for custom hook implementations.

    This class contains all inject hooks: i.e. hooks receiving events and
    possibly returning custom data for injection into pydov's execution stack.

    Inject hooks allow you to capture and intercept remote server calls,
    influencing pydov's inner workings. Use with care! If you reached this
    part of the code, it is probably wise to open an issue in Github,
    since this is most likely not what you need.

    Provides all available methods with a default implementation to do
    nothing. This allows for hook subclasses to only implement the events
    they need.

    """

    def inject_meta_response(self, url):
        """Inject a response for a metadata request.

        This allows to intercept a metadata request and return a response of
        your choice.

        When at least one registered hook returns a response for a given URL,
        the remote call is not executed and instead the response from the
        last registered hook (that is non-null) is used instead.

        Metadata calls include amongst others: WFS GetCapabilities, requests
        for MD_Metadata, FC_FeatureCatalogue and XSD schemas.

        These are all calls except for WFS GetFeature requests and XML
        downloads of DOV data - these are other hooks.

        Parameters
        ----------
        url : str
            URL of the metadata request.

        Returns
        -------
        bytes, optional
            The response to use in favor of resolving the URL. Return None to
            disable this inject hook.

        """
        return None

    def inject_wfs_getfeature_response(self, query):
        """Inject a response for a WFS GetFeature request.

        This allows to intercept a WFS GetFeature request and return a
        response of your choice.

        When at least one registered hook returns a response for a given query,
        the remote call is not executed and instead the response from the
        last registered hook (that is non-null) is used instead.

        Because of parallel processing, this method will be called
        simultaneously from multiple threads. Make sure your implementation is
        threadsafe or uses locking.

        Parameters
        ----------
        query : etree.ElementTree
            The WFS GetFeature request sent to the WFS server.

        Returns
        -------
        xml: bytes, optional
            The GetFeature response to use in favor of resolving the URL.
            Return None to disable this inject hook.

        """
        return None

    def inject_xml_response(self, pkey_object):
        """Inject a response for a DOV XML request.

        This allows to intercept a DOV XML request and return a response of
        your choice.

        When at least one registered hook returns a response for a given pkey,
        the remote call is not executed and instead the response from the
        last registered hook (that is non-null) is used instead.

        Because of parallel processing, this method will be called
        simultaneously from multiple threads. Make sure your implementation is
        threadsafe or uses locking.

        Parameters
        ----------
        query : etree.ElementTree
            The WFS GetFeature request sent to the WFS server.

        Returns
        -------
        xml : bytes, optional
            The XML response to use in favor of resolving the URL. Return
            None to disable this inject hook.

        """
        return None


class SimpleStatusHook(AbstractReadHook):
    """Simple hook implementation to print progress to stdout."""

    class ProgressState(object):
        """Simple class for storing progress state.
        """

        def __init__(self):
            """Initialise a new progress state, with all variable to default.
            """
            self.reset()

        def reset(self):
            """Reset the progress state. Resets all variables to default."""
            self.max_results = None
            self.result_count = 0
            self.prog_counter = 0
            self.init_time = time.time()
            self.previous_remaining = None

    def __init__(self):
        """Initialisation.

        Initialise all variables to default.

        """
        self.wfs_progress = SimpleStatusHook.ProgressState()
        self.xml_progress = SimpleStatusHook.ProgressState()
        self.lock = Lock()

    def _write_progress(self, state, char):
        """Write progress to standard output.

        Progress is grouped on lines per 50 items, adding ``char`` for every
        item processed.

        Parameters
        ----------
        state : ProgressState
            State of current progress.
        char : str
            Single character to print.

        """
        if state.prog_counter == 0:
            sys.stdout.write('[{:03d}/{:03d}] '.format(
                state.prog_counter, state.result_count))
            sys.stdout.flush()
        elif state.prog_counter % 50 == 0:
            time_elapsed = time.time() - state.init_time
            time_per_item = time_elapsed / state.prog_counter
            remaining_mins = int((time_per_item * (
                state.result_count - state.prog_counter)) / 60)
            if remaining_mins > 1 and remaining_mins != \
                    state.previous_remaining:
                remaining = " ({:d} min. left)".format(remaining_mins)
                state.previous_remaining = remaining_mins
            else:
                remaining = ""
            sys.stdout.write('{}\n[{:03d}/{:03d}] '.format(
                remaining, state.prog_counter, state.result_count))
            sys.stdout.flush()

        sys.stdout.write(char)
        sys.stdout.flush()
        state.prog_counter += 1

        if state.prog_counter == state.result_count:
            sys.stdout.write('\n')
            sys.stdout.flush()

    def wfs_search_init(self, params):
        """When a new WFS search is started, reset all counters to 0 and set
        the maximum requested results.

        Parameters
        ----------
        params : dict
            Parameters used to initiate WFS search. These include:
                typename : str
                    Typename in the WFS service to query.
                location : pydov.util.location.AbstractLocationFilter or None
                    Location filter limiting the features to retrieve.
                filter : str (xml) or None
                    Attribute filter limiting the features to retrieve.
                sort_by : str (xml) or None
                    SortBy clause listing fields to sort by.
                max_features : int or None
                    Limit the maximum number of features to request.
                propertynames : list of str
                    List of WFS propertynames (attributes) to retrieve.
                geometry_column : str
                    Name of the column/attribute containing the geometry.

        """
        self.wfs_progress.reset()
        self.xml_progress.reset()

        self.wfs_progress.max_results = params.get('max_features', None)
        self.xml_progress.max_results = params.get('max_features', None)

    def wfs_search_result(self, number_matched, number_returned):
        """When the WFS search completes, set the total result count.

        Parameters
        ----------
        number_matched : int
            The number of features matched by the WFS search query.
        number_returned : int
            The number of features returned by the WFS search query. Due to
            server limitations this can be less than number_matched.

        """
        if self.wfs_progress.result_count == 0:

            if self.wfs_progress.max_results is not None:
                total_results = min(
                    self.wfs_progress.max_results, number_matched)
            else:
                total_results = number_matched

            if number_returned > 0:
                self.wfs_progress.result_count = math.ceil(
                    total_results/number_returned)
            else:
                self.wfs_progress.result_count = 0

            self.xml_progress.result_count = total_results

        self._write_progress(self.wfs_progress, '.')

    def xml_cache_hit(self, pkey_object):
        """When an XML document is retrieved from the cache, print 'c' to
        the progress output.

        Parameters
        ----------
        pkey_object : str
            Permanent key of the requested object.

        """
        with self.lock:
            self._write_progress(self.xml_progress, 'c')

    def xml_stale_hit(self, pkey_object):
        """When a stale XML document is retrieved from the cache, print 'S' to
        the progress output.

        Parameters
        ----------
        pkey_object : str
            Permanent key of the requested object.

        """
        with self.lock:
            self._write_progress(self.xml_progress, 'S')

    def xml_fetch_error(self, pkey_object):
        """When an XML document failed to be fetched from DOV, print 'E' to
        the progress output.

        Parameters
        ----------
        pkey_object : str
            Permanent key of the requested object.

        """
        with self.lock:
            self._write_progress(self.xml_progress, 'E')

    def xml_downloaded(self, pkey_object):
        """When an XML document is downloaded from the DOV services,
        print '.' to the progress output.

        Parameters
        ----------
        pkey_object : str
            Permanent key of the requested object.

        """
        with self.lock:
            self._write_progress(self.xml_progress, '.')


class RepeatableLogRecorder(AbstractReadHook, AbstractInjectHook):
    """Class for recording a pydov session into a ZIP archive.

    This enables to save (the results of) all metadata and data requests from
    the DOV services locally on disk.

    The saved ZIP archive can subsequently be used in the
    `RepeatableLogReplayer` to replay the saved session allowing fully
    reproducible pydov runs.

    """

    def __init__(self, log_directory):
        """Initialise a RepeatableLogRecorder hook.

        It will save a ZIP archive with the current pydov session's data in
        the given log directory.

        Replays previously saved responses in the current session too to
        enable full reproducibility between the session being saved and replays
        of the saved ZIP archive. When this would be omitted they would not
        necessarily yield the same results due to timing issues (albeit
        quite far fetched).

        Parameters
        ----------
        log_directory : str
            Path to a directory on disk where the ZIP archive containing the
            pydov session will be saved. Will be created if it does not exist.

        """
        self.log_directory = log_directory

        if not os.path.exists(self.log_directory):
            os.makedirs(self.log_directory)

        self.log_archive = os.path.join(
            self.log_directory,
            time.strftime('pydov-archive-%Y%m%dT%H%M%S-{}.zip'.format(
                str(uuid.uuid4())[0:6]))
        )

        self.log_archive_file = zipfile.ZipFile(
            self.log_archive, 'w', compression=zipfile.ZIP_DEFLATED)

        self.metadata = {
            'versions': dict([m, self._safe_get_version(m)] for m in [
                'pydov', 'owslib', 'pandas', 'numpy', 'requests',
                'fiona', 'geopandas']),
            'timings': {
                'start': time.strftime('%Y%m%dT%H%M%S')
            }
        }
        self.started_at = time.perf_counter()

        self._store_pydov_code()

        self.lock = Lock()
        atexit.register(self._pydov_exit)

    def _safe_get_version(self, module):
        """Try to get the version string of the given module and return None
        on error or when module is not found.

        Parameters
        ----------
        module : string
            Name of the module.

        Returns
        -------
        string
            Version string of the module.
        """
        ms = importlib.util.find_spec(module)
        if ms is not None:
            try:
                return importlib.import_module(module).__version__
            except Exception:
                return None
        else:
            return None

    def _store_pydov_code(self):
        """Store the pydov source code itself in the archive.

        To get a fully reproducible pydov run, one has to a) save and replay
        all remote DOV data and b) rerun with the same pydov version for code
        changes can effect the result too.

        One can rerun an archive with the saved code by prepending the ZIP
        archive to the system path before importing pydov::

            import sys
            sys.path.insert(0, 'C:\\pydov-archive-20200128T134936-96bda7.zip')
            import pydov

        """
        pydov_root = Path(pydov.__file__).parent
        for f in pydov_root.glob('**/*.py'):
            self.log_archive_file.write(
                str(f), 'pydov/' + str(f.relative_to(pydov_root)))

    def _pydov_exit(self):
        """Save metadata and close ZIP archive before ending Python session."""
        self.metadata['timings']['end'] = time.strftime('%Y%m%dT%H%M%S')
        self.metadata['timings']['run_time_secs'] = (
            time.perf_counter() - self.started_at)

        try:
            self.log_archive_file.writestr(
                'metadata.json', json.dumps(self.metadata, indent=2))
            self.log_archive_file.close()
        except ValueError:
            # already closed, this should only happen while running tests
            pass
        else:
            print('pydov session was saved as {}'.format(self.log_archive))

    def meta_received(self, url, response):
        """Called when a response for a metadata requests is received.

        Create a stable hash based on the URL and archive the response.

        Parameters
        ----------
        url : str
            URL of the metadata request.
        response : bytes
            The raw response as received from resolving the URL.

        """
        md5_hash = md5(url.encode('utf8')).hexdigest()
        log_path = 'meta/' + md5_hash + '.log'

        if log_path not in self.log_archive_file.namelist():
            self.log_archive_file.writestr(log_path, response.decode('utf8'))

    def inject_meta_response(self, url):
        """Inject a response for a metadata request.

        Create a stable hash based on the URL and inject a previously saved
        response if available. If no previous response is available, return
        None to resume normal pydov flow.

        Parameters
        ----------
        url : str
            URL of the metadata request.

        Returns
        -------
        bytes, optional
            The response to use in favor of resolving the URL. Returns None if
            no previously recorded response is available for this request.

        """
        md5_hash = md5(url.encode('utf8')).hexdigest()
        log_path = 'meta/' + md5_hash + '.log'

        if log_path not in self.log_archive_file.namelist():
            return None

        with self.log_archive_file.open(log_path, 'r') as log_file:
            response = log_file.read()

        return response

    def wfs_search_result_received(self, query, features):
        """Called after a WFS search finished.

        Create a stable hash based on the query and archive the feature
        response.

        Parameters
        ----------
        query : etree.ElementTree
            The WFS GetFeature request sent to the WFS server.
        features : etree.ElementTree
            The WFS GetFeature response containings the features.

        """
        with self.lock:
            q = etree.tostring(query, encoding='unicode')
            md5_hash = md5(q.encode('utf8')).hexdigest()
            log_path = 'wfs/' + md5_hash + '.log'

            if log_path not in self.log_archive_file.namelist():
                self.log_archive_file.writestr(
                    log_path,
                    etree.tostring(features, encoding='utf8').decode('utf8'))

    def inject_wfs_getfeature_response(self, query):
        """Inject a response for a WFS GetFeature request.

        Create a stable hash based on the query and inject a previously saved
        response if available. If no previous response is available, return
        None to resume normal pydov flow.

        Parameters
        ----------
        query : etree.ElementTree
            The WFS GetFeature request sent to the WFS server.

        Returns
        -------
        xml: bytes, optional
            The GetFeature response to use in favor of resolving the URL.
            Return None to disable this inject hook.

        """
        with self.lock:
            q = etree.tostring(query, encoding='unicode')
            md5_hash = md5(q.encode('utf8')).hexdigest()
            log_path = 'wfs/' + md5_hash + '.log'

            if log_path not in self.log_archive_file.namelist():
                return None

            with self.log_archive_file.open(log_path, 'r') as log_file:
                tree = log_file.read()

        return tree

    def xml_received(self, pkey_object, xml):
        """Called when the XML of a given object is received, either from
        the cache or from the remote DOV service.

        Create a stable hash based on the pkey_object and archive the xml
        response.

        Parameters
        ----------
        pkey_object : str
            Permanent key of the retrieved object.
        xml : bytes
            The raw XML data of this DOV object as bytes.

        """
        with self.lock:
            md5_hash = md5(pkey_object.encode('utf8')).hexdigest()
            log_path = 'xml/' + md5_hash + '.log'

            if log_path not in self.log_archive_file.namelist():
                self.log_archive_file.writestr(log_path, xml.decode('utf8'))

    def inject_xml_response(self, pkey_object):
        """Inject a response for a DOV XML request.

        Create a stable hash based on the pkey_object and inject a previously
        saved response if available. If no previous response is available,
        return None to resume normal pydov flow.

        Parameters
        ----------
        pkey_object : str
            The permanent key of the DOV object.

        Returns
        -------
        xml : bytes, optional
            The XML response to use in favor of resolving the URL. Return
            None to disable this inject hook.

        """
        with self.lock:
            md5_hash = md5(pkey_object.encode('utf8')).hexdigest()
            log_path = 'xml/' + md5_hash + '.log'

            if log_path not in self.log_archive_file.namelist():
                return None

            with self.log_archive_file.open(log_path, 'r') as log_file:
                xml = log_file.read()

            return xml


class RepeatableLogReplayer(AbstractInjectHook):
    """Class for replaying a saved pydov session from a ZIP archive.

    This will reroute all external requests to the saved results in the ZIP
    archive, enabling fully reproducable pydov runs.

    """

    def __init__(self, log_archive):
        """Initialise a RepeatableLogReplayer hook.

        It will reroute all external requests to the saved results in the
        given ZIP archive.

        Parameters
        ----------
        log_archive : str
            Path to the ZIP archive to use for replay.

        """
        self.log_archive = log_archive

        self.log_archive_file = zipfile.ZipFile(
            self.log_archive, 'r', compression=zipfile.ZIP_DEFLATED)

        self.lock = Lock()

        atexit.register(self._pydov_exit)

    def _pydov_exit(self):
        """Close ZIP archive before ending Python session."""
        self.log_archive_file.close()

    def inject_meta_response(self, url):
        """Inject a response for a metadata request.

        Create a stable hash based on the URL and inject the previously saved
        response.

        Parameters
        ----------
        url : str
            URL of the metadata request.

        Returns
        -------
        bytes, optional
            The response to use in favor of resolving the URL.

        Raises
        ------
        pydov.util.errors.LogReplayError
            When the response required for this URL could not be found in the
            archive. This happens for instance when replaying an archive for
            a different pydov session than the one it was saved for.

        """
        md5_hash = md5(url.encode('utf8')).hexdigest()
        log_path = 'meta/' + md5_hash + '.log'

        if log_path not in self.log_archive_file.namelist():
            raise LogReplayError(
                'Failed to replay log: no entry for '
                'meta response of {}.'.format(md5_hash)
            )

        with self.log_archive_file.open(log_path, 'r') as log_file:
            response = log_file.read()

        return response

    def inject_wfs_getfeature_response(self, query):
        """Inject a response for a WFS GetFeature request.

        Create a stable hash based on the query and inject the previously saved
        response.

        Parameters
        ----------
        query : etree.ElementTree
            The WFS GetFeature request sent to the WFS server.

        Returns
        -------
        xml: bytes, optional
            The GetFeature response to use in favor of resolving the URL.

        Raises
        ------
        pydov.util.errors.LogReplayError
            When the response required for this URL could not be found in the
            archive. This happens for instance when replaying an archive for
            a different pydov session than the one it was saved for.

        """
        q = etree.tostring(query, encoding='unicode')
        md5_hash = md5(q.encode('utf8')).hexdigest()
        log_path = 'wfs/' + md5_hash + '.log'

        if log_path not in self.log_archive_file.namelist():
            raise LogReplayError(
                'Failed to replay log: no entry for '
                'WFS result of {}.'.format(md5_hash)
            )

        with self.log_archive_file.open(log_path, 'r') as log_file:
            tree = log_file.read()

        return tree

    def inject_xml_response(self, pkey_object):
        """Inject a response for a DOV XML request.

        Create a stable hash based on the pkey_object and inject the previously
        saved response.

        Parameters
        ----------
        pkey_object : str
            The permanent key of the DOV object.

        Returns
        -------
        xml : bytes, optional
            The XML response to use in favor of resolving the URL. Return
            None to disable this inject hook.

        Raises
        ------
        pydov.util.errors.LogReplayError
            When the response required for this URL could not be found in the
            archive. This happens for instance when replaying an archive for
            a different pydov session than the one it was saved for.

        """
        with self.lock:
            md5_hash = md5(pkey_object.encode('utf8')).hexdigest()
            log_path = 'xml/' + md5_hash + '.log'

            if log_path not in self.log_archive_file.namelist():
                raise LogReplayError(
                    'Failed to replay log: no entry for '
                    'XML result of {}.'.format(md5_hash)
                )

            with self.log_archive_file.open(log_path, 'r') as log_file:
                xml = log_file.read()

            return xml



"""Module grouping tests for the interpretaties search module."""

import pandas as pd
from owslib.fes2 import PropertyIsEqualTo

from pydov.search.interpretaties import LithologischeBeschrijvingenSearch
from pydov.types.fields import ReturnFieldList
from pydov.types.interpretaties import LithologischeBeschrijvingen
from tests.abstract import AbstractTestSearch

location_md_metadata = \
    'tests/data/types/interpretaties/lithologische_beschrijvingen/' \
    'md_metadata.xml'
location_fc_featurecatalogue = \
    'tests/data/types/interpretaties/lithologische_beschrijvingen/' \
    'fc_featurecatalogue.xml'
location_wfs_describefeaturetype = \
    'tests/data/types/interpretaties/lithologische_beschrijvingen/' \
    'wfsdescribefeaturetype.xml'
location_wfs_getfeature = \
    'tests/data/types/interpretaties/lithologische_beschrijvingen/' \
    'wfsgetfeature.xml'
location_wfs_feature = \
    'tests/data/types/interpretaties/lithologische_beschrijvingen/feature.xml'
location_dov_xml = \
    'tests/data/types/interpretaties/lithologische_beschrijvingen' \
    '/lithologische_beschrijvingen.xml'
location_xsd_base = \
    'tests/data/types/interpretaties/lithologische_beschrijvingen/xsd_*.xml'


class TestLithologischeBeschrijvingenSearch(AbstractTestSearch):

    search_instance = LithologischeBeschrijvingenSearch()
    datatype_class = LithologischeBeschrijvingen

    valid_query_single = PropertyIsEqualTo(propertyname='Proefnummer',
                                           literal='kb15d28w-B345')

    inexistent_field = 'onbestaand'
    wfs_field = 'Proefnummer'
    xml_field = 'beschrijving'

    valid_returnfields = ReturnFieldList.from_field_names('pkey_interpretatie',
                                                          'betrouwbaarheid_interpretatie')
    valid_returnfields_subtype = ReturnFieldList.from_field_names(
        'pkey_interpretatie', 'diepte_laag_van', 'diepte_laag_tot')
    valid_returnfields_extra = ReturnFieldList.from_field_names('pkey_interpretatie', 'gemeente')

    df_default_columns = [
        'pkey_interpretatie', 'pkey_boring',
        'betrouwbaarheid_interpretatie', 'x', 'y',
        'start_interpretatie_mtaw',
        'diepte_laag_van', 'diepte_laag_tot', 'beschrijving']

    def test_search_customreturnfields(self, mp_get_schema,
                                       mp_remote_describefeaturetype,
                                       mp_remote_wfs_feature, mp_dov_xml):
        """Test the search method with custom return fields.

        Test whether the output dataframe is correct.

        Parameters
        ----------
        mp_get_schema : pytest.fixture
            Monkeypatch the call to a remote OWSLib schema.
        mp_remote_describefeaturetype : pytest.fixture
            Monkeypatch the call to a remote DescribeFeatureType .
        mp_remote_wfs_feature : pytest.fixture
            Monkeypatch the call to get WFS features.
        mp_dov_xml : pytest.fixture
            Monkeypatch the call to get the remote XML data.

        """
        df = self.search_instance.search(
            query=self.valid_query_single,
            return_fields=('pkey_interpretatie', 'pkey_boring')
        )

        assert list(df) == ['pkey_interpretatie', 'pkey_boring']

        assert not pd.isnull(df.pkey_boring[0])

    def test_search_xml_resolve(self, mp_get_schema,
                                mp_remote_describefeaturetype,
                                mp_remote_wfs_feature, mp_dov_xml):
        """Test the search method with return fields from XML but not from a
        subtype.

        Test whether the output dataframe contains the resolved XML data.

        Parameters
        ----------
        mp_get_schema : pytest.fixture
            Monkeypatch the call to a remote OWSLib schema.
        mp_remote_describefeaturetype : pytest.fixture
            Monkeypatch the call to a remote DescribeFeatureType.
        mp_remote_wfs_feature : pytest.fixture
            Monkeypatch the call to get WFS features.
        mp_dov_xml : pytest.fixture
            Monkeypatch the call to get the remote XML data.

        """
        df = self.search_instance.search(
            query=self.valid_query_single,
            return_fields=('pkey_interpretatie', 'diepte_laag_tot'))

        assert df.diepte_laag_tot[0] == 1



"""Module grouping tests for the bodemclassificatie search module."""
from owslib.fes2 import PropertyIsEqualTo
from shapely.geometry import Point

from pydov.search.bodemclassificatie import BodemclassificatieSearch
from pydov.types.bodemclassificatie import Bodemclassificatie
from pydov.types.fields import ReturnFieldList
from pydov.util.dovutil import build_dov_url
from tests.abstract import AbstractTestSearch

location_md_metadata = 'tests/data/types/bodemclassificatie/md_metadata.xml'
location_fc_featurecatalogue = \
    'tests/data/types/bodemclassificatie/fc_featurecatalogue.xml'
location_wfs_describefeaturetype = \
    'tests/data/types/bodemclassificatie/wfsdescribefeaturetype.xml'
location_wfs_getfeature = 'tests/data/types/bodemclassificatie/wfsgetfeature.xml'
location_wfs_feature = 'tests/data/types/bodemclassificatie/feature.xml'
location_dov_xml = 'tests/data/types/bodemclassificatie/bodemclassificatie.xml'
location_xsd_base = 'tests/data/types/bodemclassificatie/xsd_*.xml'


class TestBodemclassificatieSearch(AbstractTestSearch):

    search_instance = BodemclassificatieSearch()
    datatype_class = Bodemclassificatie

    valid_query_single = PropertyIsEqualTo(
        propertyname='pkey_bodemclassificatie',
        literal=build_dov_url('data/belgischebodemclassificatie/2018-000146'))

    inexistent_field = 'onbestaand'
    wfs_field = 'bodemtype'
    xml_field = None

    valid_returnfields = ReturnFieldList.from_field_names('pkey_bodemclassificatie', 'bodemtype',
                                                          'classificatietype')
    valid_returnfields_subtype = None
    valid_returnfields_extra = None

    df_default_columns = [
        'pkey_bodemclassificatie', 'pkey_bodemlocatie', 'x', 'y', 'mv_mtaw',
        'classificatietype', 'bodemtype', 'auteurs'
    ]



"""Module grouping tests for the pydov.types.boring module."""
from pydov.types.fields import ReturnFieldList
from pydov.types.grondwatermonster import GrondwaterMonster
from pydov.util.dovutil import build_dov_url
from tests.abstract import AbstractTestTypes

location_wfs_getfeature = 'tests/data/types/grondwatermonster/' \
    'wfsgetfeature.xml'
location_wfs_feature = 'tests/data/types/grondwatermonster/feature.xml'
location_dov_xml = 'tests/data/types/grondwatermonster/grondwatermonster.xml'


class TestGrondwaterMonster(AbstractTestTypes):
    """Class grouping tests for the
    pydov.types.grondwaterfilter.GrondwaterFilter class."""

    datatype_class = GrondwaterMonster
    namespace = 'http://dov.vlaanderen.be/grondwater/gw_meetnetten'
    pkey_base = build_dov_url('data/watermonster/')

    field_names = [
        'pkey_grondwatermonster', 'grondwatermonsternummer',
        'pkey_grondwaterlocatie', 'gw_id', 'pkey_filter',
        'filternummer', 'x', 'y', 'start_grondwaterlocatie_mtaw',
        'gemeente', 'datum_monstername', 'parametergroep',
        'parameter', 'detectie', 'waarde', 'eenheid', 'veld_labo']
    field_names_subtypes = [
        'parametergroep', 'parameter', 'detectie',
        'waarde', 'eenheid', 'veld_labo']
    field_names_nosubtypes = [
        'pkey_grondwatermonster', 'grondwatermonsternummer',
        'pkey_grondwaterlocatie', 'gw_id', 'pkey_filter',
        'filternummer', 'x', 'y', 'start_grondwaterlocatie_mtaw',
        'gemeente', 'datum_monstername']

    valid_returnfields = ReturnFieldList.from_field_names('y', 'gemeente')
    valid_returnfields_subtype = ReturnFieldList.from_field_names(
        'pkey_filter', 'pkey_grondwatermonster', 'eenheid')

    inexistent_field = 'onbestaand'



# -*- coding: utf-8 -*-
"""Module grouping all classes related to pydov field definitions."""


class XsdType(object):
    """Class for specifying an XSD type from an XSD schema. This will be
    resolved at runtime in a list of possible values and their definitions."""

    def __init__(self, xsd_schema, typename):
        """Initialise a XSD type reference.

        Parameters
        ----------
        xsd_schema : str
            URL of XSD schema record containing the specified typename.
        typename : str
            Name of the type.

        """
        self.xsd_schema = xsd_schema
        self.typename = typename


class AbstractField(dict):
    """Abstract base class for pydov field definitions. Not to be
    instantiated directly."""

    def __init__(self, name, source, datatype, **kwargs):
        """Initialise a field.

        Parameters
        ----------
        name : str
            Name of this field in the return dataframe.
        source : one of 'wfs', 'xml', 'custom_wfs', 'custom_xml'
            Source of this field.
        datatype : one of 'string', 'integer', 'float', 'date', 'datetime' \
                   or 'boolean'
            Datatype of the values of this field in the return dataframe.

        """
        super(AbstractField, self).__init__(**kwargs)
        self.__setitem__('name', name)
        self.__setitem__('source', source)
        self.__setitem__('type', datatype)


class WfsField(AbstractField):
    """Class for a field available in the WFS service."""

    def __init__(self, name, source_field, datatype):
        """Initialise a WFS field.

        Parameters
        ----------
        name : str
            Name of this field in the return dataframe.
        source_field : str
            Name of this field in the source WFS service.
        datatype : one of 'string', 'integer', 'float', 'date', 'datetime' \
                   or 'boolean'
            Datatype of the values of this field in the return dataframe.

        """
        super(WfsField, self).__init__(name, 'wfs', datatype)
        self.__setitem__('sourcefield', source_field)


class _WfsInjectedField(WfsField):
    """Class for a field available in the WFS service, but not included in the
    default dataframe output."""

    def __init__(self, name, datatype):
        """Initialise a WFS injected field.

        This is a field not normally present in the dataframe, but usable as
        a query and returnfield as it is available in the WFS service.

        Parameters
        ----------
        name : str
            Name of this field in the return dataframe.
        datatype : one of 'string', 'integer', 'float', 'date', 'datetime' \
                   or 'boolean'
            Datatype of the values of this field in the return dataframe.

        """
        super(_WfsInjectedField, self).__init__(name, name, datatype)
        self.__setitem__('wfs_injected', True)


class XmlField(AbstractField):
    """Class for a field available in the XML document."""

    def __init__(self, name, source_xpath, datatype, definition='',
                 notnull=False, xsd_type=None):
        """Initialise an XML field.

        Parameters
        ----------
        name : str
            Name of this field in the return dataframe.
        source_xpath : str
            XPath expression of the values of this field in the source XML
            document.
        datatype : one of 'string', 'integer', 'float', 'date', 'datetime' \
                   or 'boolean'
            Datatype of the values of this field in the return dataframe.
        definition : str, optional
            Definition of this field.
        notnull : bool, optional, defaults to False
            True if this field is always present (mandatory), False otherwise.
        xsd_type : pydov.types.abstract.XsdType, optional
            XSD type associated with this field.

        """
        super(XmlField, self).__init__(name, 'xml', datatype)

        self.__setitem__('sourcefield', source_xpath)
        self.__setitem__('definition', definition)
        self.__setitem__('notnull', notnull)

        if xsd_type is not None:
            self.__setitem__('xsd_schema', xsd_type.xsd_schema)
            self.__setitem__('xsd_type', xsd_type.typename)


class _CustomWfsField(AbstractField):
    """Class for a custom field, created explicitly in pydov from other WFS
    fields."""

    def __init__(self, name, datatype, definition='', notnull=False):
        """Initialise a custom field.

        Parameters
        ----------
        name : str
            Name of this field in the return dataframe.
        datatype : one of 'string', 'integer', 'float', 'date', 'datetime' \
                   or 'boolean'
            Datatype of the values of this field in the return dataframe.
        definition : str, optional
            Definition of this field.
        notnull : bool, optional, defaults to False
            True if this field is always present (mandatory), False otherwise.

        """
        super(_CustomWfsField, self).__init__(name, 'custom_wfs', datatype)
        self.__setitem__('definition', definition)
        self.__setitem__('notnull', notnull)

    def requires_wfs_fields(self):
        """Get a list of WFS fields that are required by (the calculation of)
        this custom field.

        Returns
        -------
        list of str
            List of WFS fieldnames that is required by this custom field.

        Raises
        ------
        NotImplementedError
            Implement this in a subclass.
        """
        raise NotImplementedError

    def calculate(self, instance):
        """Calculate the value of this custom field for the given instance.

        Parameters
        ----------
        instance : AbstractDovType
            Instance of the corresponding type, containing all WFS values in
            its data dictionary.

        Returns
        -------
        Value to be used for this custom field for this instance. Its datatype
        should match the one set in the initialisation of the custom field.

        Raises
        ------
        NotImplementedError
            Implement this in a subclass.
        """
        raise NotImplementedError


class _CustomXmlField(AbstractField):
    """Class for a custom field, created explicitly in pydov from other XML
    fields."""

    def __init__(self, name, datatype, definition='', notnull=False):
        """Initialise a custom field.

        Parameters
        ----------
        name : str
            Name of this field in the return dataframe.
        datatype : one of 'string', 'integer', 'float', 'date', 'datetime' \
                   or 'boolean'
            Datatype of the values of this field in the return dataframe.
        definition : str, optional
            Definition of this field.
        notnull : bool, optional, defaults to False
            True if this field is always present (mandatory), False otherwise.

        """
        super(_CustomXmlField, self).__init__(name, 'custom_xml', datatype)
        self.__setitem__('definition', definition)
        self.__setitem__('notnull', notnull)

    def calculate(self, cls, tree):
        """Calculate the value of this custom field from the given XML tree.

        Parameters
        ----------
        cls : AbstractDovType
            Class of the type this field belongs to.
        tree : etree.ElementTree
            ElementTree of the DOV XML for this instance.

        Returns
        -------
        Value to be used for this custom field for this instance. Its datatype
        should match the one set in the initialisation of the custom field.

        Raises
        ------
        NotImplementedError
            Implement this in a subclass.
        """
        raise NotImplementedError


class ReturnFieldList(list):
    """List of return fields used in search methods. """

    def __contains__(self, __key: object):
        """Overwrite the default method. Checks on field name.

        Parameters
        ----------
        __key
            The key to check.

        Returns
        -------
        boolean
            Whether the key is one of the names in the ReturnFieldList.
        """
        return __key in [i.name for i in self]

    def get_names(self):
        """Return a list with all the names of the fields in this
        ReturnFieldList.

        Returns
        -------
        list of str
            List of field names.
        """
        return [f.name for f in self]

    @classmethod
    def from_field_names(self, *return_fields):
        """Initiale a ReturnFieldList from a list of return field names.

        Parameters
        ----------
        return_fields : list or set or tuple of str
            List, set or tuple of return field names.

        Returns
        -------
        ReturnFieldList
            Equivalent ReturnFieldList.

        Raises
        ------
        AttributeError
            If the value of return_fields is not a list, set or tuple.
        """
        if len(return_fields) == 0:
            return None

        if len(return_fields) == 1:
            return_fields = return_fields[0]

        if isinstance(return_fields, str):
            return_fields = (return_fields,)

        if return_fields is None:
            return None

        if not isinstance(return_fields, (list, set, tuple)):
            # FIXME: this should be TypeError instead
            raise AttributeError(
                'return_fields should be a list, set or tuple')

        result = ReturnFieldList()
        for rf in return_fields:
            if isinstance(rf, AbstractReturnField):
                result.append(rf)
            else:
                result.append(ReturnField.from_field_name(rf))
        return result


class AbstractReturnField:
    """Base class of ReturnField and GeometryReturnField."""

    def __init__(self, name):
        """Initialisation.

        Parameters
        ----------
        name : str
            Name of the return field.
        """
        self.name = name

    @classmethod
    def from_field_name(cls, name):
        """Initialise a new instance from a field name.

        Parameters
        ----------
        name : str
            Field name.

        Returns
        -------
        Instance of this class.
            Instance of ReturnField or GeometryReturnField.
        """
        return cls(name)


class ReturnField(AbstractReturnField):
    """Normal (non-geometry) return field."""

    def __init__(self, name):
        """Initialisation.

        Parameters
        ----------
        name : str
            Name of the return field.
        """
        super().__init__(name)


class GeometryReturnField(AbstractReturnField):
    def __init__(self, geometry_field, epsg=None):
        """Initialise a geometry return field.

        Parameters
        ----------
        geometry_field : str
            Name of the geometry field.
        epsg : int, optional
            EPSG code of the CRS of the geometries that will be returned.
            Defaults to None, which means the default CRS of the WFS layer.
        """
        super().__init__(geometry_field)

        if epsg is not None:
            if not isinstance(epsg, int):
                raise TypeError('epsg should be an integer value')

        self.epsg = epsg



# -*- coding: utf-8 -*-
"""Module containing the DOV data types for interpretations, including
subtypes."""

from pydov.types.abstract import AbstractDovSubType, AbstractDovType
from pydov.types.fields import WfsField, XmlField, XsdType, _CustomWfsField
from pydov.util.dovutil import build_dov_url


class PkeyBoringField(_CustomWfsField):
    """Custom field to populate pkey_boring in case the interpretation is
    linked with a Boring.
    """

    def __init__(self):
        super().__init__(name='pkey_boring',
                         definition='URL die verwijst naar de gegevens van de '
                         'boring waaraan deze interpretatie '
                         'gekoppeld is (indien gekoppeld aan een '
                         'boring).',
                         datatype='string')

    def requires_wfs_fields(self):
        return ['Type_proef', 'Proeffiche']

    def calculate(self, instance):
        if instance.data['Type_proef'] == 'Boring':
            return instance.data['Proeffiche']


class PkeySonderingField(_CustomWfsField):
    """Custom field to populate pkey_sondering in case the interpretation is
    linked with a Sondering.
    """

    def __init__(self):
        super().__init__(name='pkey_sondering',
                         definition='URL die verwijst naar de gegevens van de '
                         'sondering waaraan deze interpretatie '
                         'gekoppeld is (indien gekoppeld '
                         'aan een sondering).',
                         datatype='string')

    def requires_wfs_fields(self):
        return ['Type_proef', 'Proeffiche']

    def calculate(self, instance):
        if instance.data['Type_proef'] == 'Sondering':
            return instance.data['Proeffiche']


class AbstractCommonInterpretatie(AbstractDovType):
    """Abstract base class for interpretations that can be linked to
    boreholes or cone penetration tests."""

    fields = [
        WfsField(name='pkey_interpretatie',
                 source_field='Interpretatiefiche', datatype='string'),
        PkeyBoringField(),
        PkeySonderingField(),
        WfsField(name='betrouwbaarheid_interpretatie',
                 source_field='Betrouwbaarheid', datatype='string'),
        WfsField(name='x', source_field='X_mL72', datatype='float'),
        WfsField(name='y', source_field='Y_mL72', datatype='float'),
        WfsField(name='start_interpretatie_mtaw', source_field='Z_mTAW',
                 datatype='float')
    ]

    pkey_fieldname = 'Interpretatiefiche'

    def __init__(self, pkey):
        """Initialisation.

        Parameters
        ----------
        pkey : str
            Permanent key of the Interpretatie (interpretations), being a
            URI of the form
            `https://www.dov.vlaanderen.be/data/interpretatie/<id>`.

        """
        super().__init__('interpretatie', pkey)


class AbstractBoringInterpretatie(AbstractDovType):
    """Abstract base class for interpretations that are linked to boreholes
    only."""

    fields = [
        WfsField(name='pkey_interpretatie',
                 source_field='Interpretatiefiche', datatype='string'),
        WfsField(name='pkey_boring', source_field='Proeffiche',
                 datatype='string'),
        WfsField(name='betrouwbaarheid_interpretatie',
                 source_field='Betrouwbaarheid', datatype='string'),
        WfsField(name='x', source_field='X_mL72', datatype='float'),
        WfsField(name='y', source_field='Y_mL72', datatype='float'),
        WfsField(name='start_interpretatie_mtaw', source_field='Z_mTAW',
                 datatype='float')
    ]

    pkey_fieldname = 'Interpretatiefiche'

    def __init__(self, pkey):
        """Initialisation.

        Parameters
        ----------
        pkey : str
            Permanent key of the Interpretatie (interpretations), being a
            URI of the form
            `https://www.dov.vlaanderen.be/data/interpretatie/<id>`.

        """
        super().__init__('interpretatie', pkey)

    @classmethod
    def from_wfs_element(cls, feature, namespace):
        instance = cls(
            feature.findtext('./{{{}}}{}'.format(
                namespace, cls.pkey_fieldname)))

        for field in cls.get_fields(source=('wfs',)).values():
            instance.data[field['name']] = cls._parse(
                func=feature.findtext,
                xpath=field['sourcefield'],
                namespace=namespace,
                returntype=field.get('type', None)
            )

        return instance


class InformeleStratigrafieLaag(AbstractDovSubType):

    rootpath = './/informelestratigrafie/laag'

    fields = [
        XmlField(name='diepte_laag_van',
                 source_xpath='/van',
                 definition='Diepte van de bovenkant van de laag informele '
                            'stratigrafie in meter.',
                 datatype='float'),
        XmlField(name='diepte_laag_tot',
                 source_xpath='/tot',
                 definition='Diepte van de onderkant van de laag informele '
                            'stratigrafie in meter.',
                 datatype='float'),
        XmlField(name='beschrijving',
                 source_xpath='/beschrijving',
                 definition='Benoeming van de eenheid van de laag informele '
                            'stratigrafie in vrije tekst (onbeperkt in '
                            'lengte).',
                 datatype='string')
    ]


class InformeleStratigrafie(AbstractCommonInterpretatie):
    """Class representing the DOV data type for 'informele stratigrafie'
    interpretations."""

    subtypes = [InformeleStratigrafieLaag]


class FormeleStratigrafieLaag(AbstractDovSubType):

    rootpath = './/formelestratigrafie/laag'

    fields = [
        XmlField(name='diepte_laag_van',
                 source_xpath='/van',
                 definition='Diepte van de bovenkant van de laag Formele '
                            'stratigrafie in meter.',
                 datatype='float'),
        XmlField(name='diepte_laag_tot',
                 source_xpath='/tot',
                 definition='Diepte van de onderkant van de laag Formele '
                            'stratigrafie in meter.',
                 datatype='float'),
        XmlField(name='lid1',
                 source_xpath='/lid1',
                 definition='eerste eenheid van de laag formele stratigrafie',
                 datatype='string',
                 xsd_type=XsdType(
                     xsd_schema=build_dov_url(
                         'xdov/schema/latest/xsd/kern/interpretatie/'
                         'FormeleStratigrafieDataCodes.xsd'),
                     typename='FormeleStratigrafieLedenEnumType')),
        XmlField(name='relatie_lid1_lid2',
                 source_xpath='/relatie_lid1_lid2',
                 definition='verbinding/relatie tussen lid1 en lid2 van de '
                            'laag formele stratigrafie',
                 datatype='string',
                 xsd_type=XsdType(
                     xsd_schema=build_dov_url(
                         'xdov/schema/latest/xsd/kern/interpretatie/'
                         'InterpretatieDataCodes.xsd'),
                     typename='RelatieLedenEnumType')),
        XmlField(name='lid2',
                 source_xpath='/lid2',
                 definition='tweede eenheid van de laag formele stratigrafie. '
                      'Indien niet ingevuld wordt default de waarde van lid1 '
                      'ingevuld',
                 datatype='string',
                 xsd_type=XsdType(
                     xsd_schema=build_dov_url(
                         'xdov/schema/latest/xsd/kern/interpretatie/'
                         'FormeleStratigrafieDataCodes.xsd'),
                     typename='FormeleStratigrafieLedenEnumType'))
    ]


class FormeleStratigrafie(AbstractCommonInterpretatie):
    """Class representing the DOV data type for 'Formele stratigrafie'
    interpretations."""

    subtypes = [FormeleStratigrafieLaag]


class HydrogeologischeStratigrafieLaag(AbstractDovSubType):

    rootpath = './/hydrogeologischeinterpretatie/laag'

    fields = [
        XmlField(name='diepte_laag_van',
                 source_xpath='/van',
                 definition='Diepte van de bovenkant van de laag '
                            'hydrogeologische stratigrafie in meter.',
                 datatype='float'),
        XmlField(name='diepte_laag_tot',
                 source_xpath='/tot',
                 definition='Diepte van de onderkant van de laag '
                            'hydrogeologische stratigrafie in meter.',
                 datatype='float'),
        XmlField(name='aquifer',
                 source_xpath='/aquifer',
                 definition='Code van de watervoerende laag waarin de laag '
                            'Hydrogeologische stratigrafie zich bevindt '
                            '(HCOVv1)',
                 datatype='string',
                 xsd_type=XsdType(
                     xsd_schema=build_dov_url(
                         'xdov/schema/latest/xsd/kern/interpretatie/'
                         'HydrogeologischeStratigrafieDataCodes.xsd'),
                     typename='AquiferHCOVv1EnumType'
                 ))
    ]


class HydrogeologischeStratigrafie(AbstractBoringInterpretatie):
    """Class representing the DOV data type for 'hydrogeologische
    stratigrafie' interpretations."""

    subtypes = [HydrogeologischeStratigrafieLaag]


class LithologischeBeschrijvingLaag(AbstractDovSubType):

    rootpath = './/lithologischebeschrijving/laag'

    fields = [
        XmlField(name='diepte_laag_van',
                 source_xpath='/van',
                 definition='Diepte van de bovenkant van de laag '
                            'lithologische beschrijving in meter.',
                 datatype='float'),
        XmlField(name='diepte_laag_tot',
                 source_xpath='/tot',
                 definition='Diepte van de onderkant van de laag '
                            'lithologische beschrijving in meter.',
                 datatype='float'),
        XmlField(name='beschrijving',
                 source_xpath='/beschrijving',
                 definition='Lithologische beschrijving van de laag in vrije '
                            'tekst (onbeperkt in lengte)',
                 datatype='string')
    ]


class LithologischeBeschrijvingen(AbstractBoringInterpretatie):
    """Class representing the DOV data type for 'lithologische
    beschrijvingen' interpretations."""

    subtypes = [LithologischeBeschrijvingLaag]


class GecodeerdeLithologieLaag(AbstractDovSubType):

    rootpath = './/gecodeerdelithologie/laag'

    __gecodeerdHoofdnaamCodesEnumType = XsdType(
        xsd_schema=build_dov_url('xdov/schema/latest/xsd/kern/interpretatie/'
                                 'GecodeerdeLithologieDataCodes.xsd'),
        typename='GecodeerdHoofdnaamCodesEnumType'
    )

    __gecodeerdBijmengingHoeveelheidEnumType = XsdType(
        xsd_schema=build_dov_url(
            'xdov/schema/latest/xsd/kern/interpretatie/'
            'GecodeerdeLithologieDataCodes.xsd'),
        typename='GecodeerdBijmengingHoeveelheidEnumType'
    )

    fields = [
        XmlField(name='diepte_laag_van',
                 source_xpath='/van',
                 definition='Diepte van de bovenkant van de laag '
                            'gecodeerde lithologie in meter.',
                 datatype='float'),
        XmlField(name='diepte_laag_tot',
                 source_xpath='/tot',
                 definition='Diepte van de onderkant van de laag '
                            'gecodeerde lithologie in meter.',
                 datatype='float'),
        XmlField(name='hoofdnaam1_grondsoort',
                 source_xpath='/hoofdnaam[1]/grondsoort',
                 definition='Primaire grondsoort (als code) van de laag '
                            'gecodeerde lithologie',
                 datatype='string',
                 xsd_type=__gecodeerdHoofdnaamCodesEnumType),
        XmlField(name='hoofdnaam2_grondsoort',
                 source_xpath='/hoofdnaam[2]/grondsoort',
                 definition='Secundaire grondsoort (als code) van de laag '
                            'gecodeerde lithologie',
                 datatype='string',
                 xsd_type=__gecodeerdHoofdnaamCodesEnumType),
        XmlField(name='bijmenging1_plaatselijk',
                 source_xpath='/bijmenging[1]/plaatselijk',
                 definition='plaatselijk of niet-plaatselijk',
                 datatype='boolean'),
        XmlField(name='bijmenging1_hoeveelheid',
                 source_xpath='/bijmenging[1]/hoeveelheid',
                 definition='aanduiding van de hoeveelheid bijmenging',
                 datatype='string',
                 xsd_type=__gecodeerdBijmengingHoeveelheidEnumType),
        XmlField(name='bijmenging1_grondsoort',
                 source_xpath='/bijmenging[1]/grondsoort',
                 definition='type grondsoort (als code) van de laag '
                            'gecodeerde lithologie of geotechnische codering',
                 datatype='string',
                 xsd_type=__gecodeerdHoofdnaamCodesEnumType),
        XmlField(name='bijmenging2_plaatselijk',
                 source_xpath='/bijmenging[2]/plaatselijk',
                 definition='plaatselijk of niet-plaatselijk',
                 datatype='boolean'),
        XmlField(name='bijmenging2_hoeveelheid',
                 source_xpath='/bijmenging[2]/hoeveelheid',
                 definition='aanduiding van de hoeveelheid bijmenging',
                 datatype='string',
                 xsd_type=__gecodeerdBijmengingHoeveelheidEnumType),
        XmlField(name='bijmenging2_grondsoort',
                 source_xpath='/bijmenging[2]/grondsoort',
                 definition='type grondsoort (als code) van de laag '
                            'gecodeerde lithologie of geotechnische codering',
                 datatype='string',
                 xsd_type=__gecodeerdHoofdnaamCodesEnumType),
        XmlField(name='bijmenging3_plaatselijk',
                 source_xpath='/bijmenging[3]/plaatselijk',
                 definition='plaatselijk of niet-plaatselijk',
                 datatype='boolean'),
        XmlField(name='bijmenging3_hoeveelheid',
                 source_xpath='/bijmenging[3]/hoeveelheid',
                 definition='aanduiding van de hoeveelheid bijmenging',
                 datatype='string',
                 xsd_type=__gecodeerdBijmengingHoeveelheidEnumType),
        XmlField(name='bijmenging3_grondsoort',
                 source_xpath='/bijmenging[3]/grondsoort',
                 definition='type grondsoort (als code) van de laag '
                            'gecodeerde lithologie of geotechnische codering',
                 datatype='string',
                 xsd_type=__gecodeerdHoofdnaamCodesEnumType)
    ]


class GecodeerdeLithologie(AbstractBoringInterpretatie):
    """Class representing the DOV data type for 'gecodeerde
    lithologie' interpretations."""

    subtypes = [GecodeerdeLithologieLaag]


class GeotechnischeCoderingLaag(AbstractDovSubType):

    rootpath = './/geotechnischecodering/laag'

    __geotechnischeCoderingHoofdnaamCodesEnumType = XsdType(
        xsd_schema=build_dov_url(
            'xdov/schema/latest/xsd/kern/interpretatie/'
            'GeotechnischeCoderingDataCodes.xsd'),
        typename='GeotechnischeCoderingHoofdnaamCodesEnumType'
    )

    __gtCoderingBijmengingHoeveelheidEnumType = XsdType(
        xsd_schema=build_dov_url(
            'xdov/schema/latest/xsd/kern/interpretatie/'
            'GeotechnischeCoderingDataCodes.xsd'),
        typename='GeotechnischeCoderingBijmengingHoeveelheidEnumType'
    )

    fields = [
        XmlField(name='diepte_laag_van',
                 source_xpath='/van',
                 definition='Diepte van de bovenkant van de laag '
                            'geotechnische codering in meter.',
                 datatype='float'),
        XmlField(name='diepte_laag_tot',
                 source_xpath='/tot',
                 definition='Diepte van de onderkant van de laag '
                            'geotechnische codering in meter.',
                 datatype='float'),
        XmlField(name='hoofdnaam1_grondsoort',
                 source_xpath='/hoofdnaam[1]/grondsoort',
                 definition='hoofdnaam (als code) van de laag geotechnische '
                            'codering',
                 datatype='string',
                 xsd_type=__geotechnischeCoderingHoofdnaamCodesEnumType),
        XmlField(name='hoofdnaam2_grondsoort',
                 source_xpath='/hoofdnaam[2]/grondsoort',
                 definition='Secundaire grondsoort (als code) van de laag '
                            'geotechnische codering',
                 datatype='string',
                 xsd_type=__geotechnischeCoderingHoofdnaamCodesEnumType),
        XmlField(name='bijmenging1_plaatselijk',
                 source_xpath='/bijmenging[1]/plaatselijk',
                 definition='plaatselijk of niet-plaatselijk',
                 datatype='boolean'),
        XmlField(name='bijmenging1_hoeveelheid',
                 source_xpath='/bijmenging[1]/hoeveelheid',
                 definition='aanduiding van de hoeveelheid bijmenging',
                 datatype='string',
                 xsd_type=__gtCoderingBijmengingHoeveelheidEnumType),
        XmlField(name='bijmenging1_grondsoort',
                 source_xpath='/bijmenging[1]/grondsoort',
                 definition='type grondsoort (als code) van de laag '
                            'geotechnische codering',
                 datatype='string',
                 xsd_type=__geotechnischeCoderingHoofdnaamCodesEnumType),
        XmlField(name='bijmenging2_plaatselijk',
                 source_xpath='/bijmenging[2]/plaatselijk',
                 definition='plaatselijk of niet-plaatselijk',
                 datatype='boolean'),
        XmlField(name='bijmenging2_hoeveelheid',
                 source_xpath='/bijmenging[2]/hoeveelheid',
                 definition='aanduiding van de hoeveelheid bijmenging',
                 datatype='string',
                 xsd_type=__gtCoderingBijmengingHoeveelheidEnumType),
        XmlField(name='bijmenging2_grondsoort',
                 source_xpath='/bijmenging[2]/grondsoort',
                 definition='type grondsoort (als code) van de laag '
                            'geotechnische codering',
                 datatype='string',
                 xsd_type=__geotechnischeCoderingHoofdnaamCodesEnumType),
        XmlField(name='bijmenging3_plaatselijk',
                 source_xpath='/bijmenging[3]/plaatselijk',
                 definition='plaatselijk of niet-plaatselijk',
                 datatype='boolean'),
        XmlField(name='bijmenging3_hoeveelheid',
                 source_xpath='/bijmenging[3]/hoeveelheid',
                 definition='aanduiding van de hoeveelheid bijmenging',
                 datatype='string',
                 xsd_type=__gtCoderingBijmengingHoeveelheidEnumType),
        XmlField(name='bijmenging3_grondsoort',
                 source_xpath='/bijmenging[3]/grondsoort',
                 definition='type grondsoort (als code) van de laag '
                            'geotechnische codering',
                 datatype='string',
                 xsd_type=__geotechnischeCoderingHoofdnaamCodesEnumType)
    ]


class GeotechnischeCodering(AbstractBoringInterpretatie):
    """Class representing the DOV data type for 'geotechnische
    codering' interpretations."""

    subtypes = [GeotechnischeCoderingLaag]


class QuartairStratigrafieLaag(AbstractDovSubType):

    rootpath = './/quartairstratigrafie/laag'

    fields = [
        XmlField(name='diepte_laag_van',
                 source_xpath='/van',
                 definition='diepte van de bovenkant van de laag '
                            'quartairstratigrafie in meter',
                 datatype='float'),
        XmlField(name='diepte_laag_tot',
                 source_xpath='/tot',
                 definition='diepte van de onderkant van de laag '
                            'quartairstratigrafie in meter',
                 datatype='float'),
        XmlField(name='lid1',
                 source_xpath='/lid1',
                 definition='eerste eenheid van de laag quartairstratigrafie',
                 datatype='string',
                 xsd_type=XsdType(
                     xsd_schema=build_dov_url(
                         'xdov/schema/latest/xsd/kern/interpretatie/'
                         'QuartairStratigrafieDataCodes.xsd'),
                     typename='QuartairStratigrafieLedenEnumType')),
        XmlField(name='relatie_lid1_lid2',
                 source_xpath='/relatie_lid1_lid2',
                 definition='verbinding of relatie tussen lid1 en lid2 van de '
                            'laag quartairstratigrafie',
                 datatype='string',
                 xsd_type=XsdType(
                     xsd_schema=build_dov_url(
                         'xdov/schema/latest/xsd/kern/interpretatie/'
                         'InterpretatieDataCodes.xsd'),
                     typename='RelatieLedenEnumType')),
        XmlField(name='lid2',
                 source_xpath='/lid2',
                 definition='tweede eenheid van de laag quartairstratigrafie. '
                      'Indien niet ingevuld wordt default dezelfde waarde '
                      'als voor Lid1 ingevuld',
                 datatype='string',
                 xsd_type=XsdType(
                     xsd_schema=build_dov_url(
                         'xdov/schema/latest/xsd/kern/interpretatie/'
                         'QuartairStratigrafieDataCodes.xsd'),
                     typename='QuartairStratigrafieLedenEnumType'))
    ]


class QuartairStratigrafie(AbstractBoringInterpretatie):
    """Class representing the DOV data type for 'Quartairstratigrafie'
    interpretations."""

    subtypes = [QuartairStratigrafieLaag]


class InformeleHydrogeologischeStratigrafieLaag(AbstractDovSubType):

    rootpath = './/informelehydrostratigrafie/laag'

    fields = [
        XmlField(name='diepte_laag_van',
                 source_xpath='/van',
                 definition='Diepte van de bovenkant van de laag informele '
                            'hydrostratigrafie in meter.',
                 datatype='float'),
        XmlField(name='diepte_laag_tot',
                 source_xpath='/tot',
                 definition='Diepte van de onderkant van de laag informele '
                            'hydrostratigrafie in meter.',
                 datatype='float'),
        XmlField(name='beschrijving',
                 source_xpath='/beschrijving',
                 definition='Benoeming van de eenheid van de laag informele '
                            'hydrostratigrafie in vrije tekst (onbeperkt in '
                            'lengte).',
                 datatype='string')
    ]


class InformeleHydrogeologischeStratigrafie(AbstractBoringInterpretatie):
    """Class representing the DOV data type for 'informele stratigrafie'
    interpretations."""

    subtypes = [InformeleHydrogeologischeStratigrafieLaag]



"""Class grouping tests for the fields module."""

import pytest
from pydov.types.fields import AbstractReturnField, GeometryReturnField, ReturnField, ReturnFieldList


class TestReturnFieldList():
    """Test the ReturnFieldList class."""

    def _test_from_field_names(self, field_names):
        """Helper method to test the from_field_names method.

        Parameters
        ----------
        field_names : any
            Field names used in the from_field_names method.
        """
        rfl = ReturnFieldList.from_field_names(
            field_names
        )

        assert isinstance(rfl, ReturnFieldList)

        for rf in rfl:
            assert isinstance(rf, AbstractReturnField)

        rf_names = [rf.name for rf in rfl]
        assert [f for f in field_names if f not in rf_names] == []

    def test_from_field_names_list(self):
        """Test the from_field_names method with a list of strings."""
        self._test_from_field_names(
            ['field1', 'field2']
        )

    def test_from_field_names_set(self):
        """Test the from_field_names method with a set of strings."""
        self._test_from_field_names(
            set(['field1', 'field2'])
        )

    def test_from_field_names_tuple(self):
        """Test the from_field_names method with a tuple of strings."""
        self._test_from_field_names(
            ('field1', 'field2')
        )

    def test_from_field_names_none(self):
        """Test the from_field_names method without specifying field names."""
        assert ReturnFieldList.from_field_names(None) is None
        assert ReturnFieldList.from_field_names() is None

    def test_from_field_names_parameters(self):
        """Test the from_field_names method with two string parameters."""
        rfl = ReturnFieldList.from_field_names(
            'field1', 'field2'
        )

        assert isinstance(rfl, ReturnFieldList)

        for rf in rfl:
            assert isinstance(rf, AbstractReturnField)

        rf_names = [rf.name for rf in rfl]
        assert [f for f in ('field1', 'field2') if f not in rf_names] == []

    def test_from_field_names_single_parameter(self):
        """Test the from_field_names method with a single string parameter."""
        rfl = ReturnFieldList.from_field_names(
            'field1'
        )

        assert isinstance(rfl, ReturnFieldList)

        for rf in rfl:
            assert isinstance(rf, AbstractReturnField)

        rf_names = [rf.name for rf in rfl]
        assert [f for f in ('field1',) if f not in rf_names] == []

    def test_get_names(self):
        """Test the get_names method."""
        rfl = ReturnFieldList.from_field_names('field1', 'field2')
        assert rfl.get_names() == ['field1', 'field2']

    def test_contains(self):
        """Test whether the contains check works with field names."""
        rfl = ReturnFieldList.from_field_names('field1', 'field2')

        assert 'field1' in rfl
        assert 'field3' not in rfl


class TestReturnField():
    """Class grouping tests for the ReturnField class."""

    def test_initialise(self):
        """Test initialisation of a ReturnField."""
        rf = ReturnField('field1')
        assert isinstance(rf, ReturnField)

    def test_name(self):
        """Test whether the name of the ReturnField is accessible."""
        rf = ReturnField('field1')
        assert rf.name == 'field1'


class TestGeometryReturnField():
    """Class grouping tests for the GeometryReturnField class."""

    def test_no_srs(self):
        """Test initialisation of a GeometryReturnField without an SRS."""
        rf = GeometryReturnField('shape')

        assert isinstance(rf, GeometryReturnField)
        assert rf.name == 'shape'
        assert rf.epsg is None

    def test_srs_31370(self):
        """Test initialisation of a GeometryReturnField with CRS set to Belgian Lambert 72."""
        rf = GeometryReturnField('shape', 31370)

        assert isinstance(rf, GeometryReturnField)
        assert rf.name == 'shape'
        assert rf.epsg == 31370

    def test_wrong_srs_type(self):
        """Test initialisation of a GeometryReturnField with a wrong CRS type.

        Test whether a TypeError is raised.
        """
        with pytest.raises(TypeError):
            GeometryReturnField('shape', 'EPSG:31370')
